# Copyright 2002 2005 Ben Escoto
#
# This file is part of rdiff-backup.
#
# rdiff-backup is free software; you can redistribute it and/or modify
# under the terms of the GNU General Public License as published by the
# Free Software Foundation; either version 2 of the License, or (at your
# option) any later version.
#
# rdiff-backup is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with rdiff-backup; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA
# 02110-1301, USA
"""Invoke rdiff utility to make signatures, deltas, or patch"""

from . import Globals, log, TempFile, rpath, hash, librsync


def get_signature(rp, blocksize=None):
    """Take signature of rpin file and return in file object"""
    if not blocksize:
        blocksize = find_blocksize(rp.getsize())
    log.Log(
        "Getting signature of %s with blocksize %s" % (rp.get_safeindexpath(),
                                                       blocksize), 7)
    return librsync.SigFile(rp.open("rb"), blocksize)


def find_blocksize(file_len):
    """Return a reasonable block size to use on files of length file_len

    If the block size is too big, deltas will be bigger than is
    necessary.  If the block size is too small, making deltas and
    patching can take a really long time.

    """
    if file_len < 4096:
        return 64  # set minimum of 64 bytes
    else:  # Use square root, rounding to nearest 16
        return int(pow(file_len, 0.5) / 16) * 16


def get_delta_sigfileobj(sig_fileobj, rp_new):
    """Like get_delta but signature is in a file object"""
    log.Log(
        "Getting delta of %s with signature stream" % rp_new.get_safepath(), 7)
    return librsync.DeltaFile(sig_fileobj, rp_new.open("rb"))


def get_delta_sigrp(rp_signature, rp_new):
    """Take signature rp and new rp, return delta file object"""
    log.Log(
        "Getting delta of %s with signature %s" %
        (rp_new.get_safepath(), rp_signature.get_safeindexpath()), 7)
    return librsync.DeltaFile(rp_signature.open("rb"), rp_new.open("rb"))


def get_delta_sigrp_hash(rp_signature, rp_new):
    """Like above but also calculate hash of new as close() value"""
    log.Log(
        "Getting delta (with hash) of %s with signature %s" %
        (rp_new.get_safepath(), rp_signature.get_safeindexpath()), 7)
    return librsync.DeltaFile(
        rp_signature.open("rb"), hash.FileWrapper(rp_new.open("rb")))


def write_delta(basis, new, delta, compress=None):
    """Write rdiff delta which brings basis to new"""
    log.Log(
        "Writing delta %s from %s -> %s" %
        (basis.get_safepath(), new.get_safepath(), delta.get_safepath()), 7)
    deltafile = librsync.DeltaFile(get_signature(basis), new.open("rb"))
    delta.write_from_fileobj(deltafile, compress)


def write_patched_fp(basis_fp, delta_fp, out_fp):
    """Write patched file to out_fp given input fps.  Closes input files"""
    rpath.copyfileobj(librsync.PatchedFile(basis_fp, delta_fp), out_fp)
    assert not basis_fp.close() and not delta_fp.close()


def write_via_tempfile(fp, rp):
    """Write fileobj fp to rp by writing to tempfile and renaming"""
    tf = TempFile.new(rp)
    retval = tf.write_from_fileobj(fp)
    rpath.rename(tf, rp)
    return retval


def patch_local(rp_basis, rp_delta, outrp=None, delta_compressed=None):
    """Patch routine that must be run locally, writes to outrp

    This should be run local to rp_basis because it needs to be a real
    file (librsync may need to seek around in it).  If outrp is None,
    patch rp_basis instead.

    The return value is the close value of the delta, so it can be
    used to produce hashes.

    """
    assert rp_basis.conn is Globals.local_connection
    if delta_compressed:
        deltafile = rp_delta.open("rb", 1)
    else:
        deltafile = rp_delta.open("rb")
    patchfile = librsync.PatchedFile(rp_basis.open("rb"), deltafile)
    if outrp:
        return outrp.write_from_fileobj(patchfile)
    else:
        return write_via_tempfile(patchfile, rp_basis)


def copy_local(rpin, rpout, rpnew=None):
    """Write rpnew == rpin using rpout as basis.  rpout and rpnew local"""
    assert rpout.conn is Globals.local_connection
    deltafile = rpin.conn.librsync.DeltaFile(
        get_signature(rpout), rpin.open("rb"))
    patched_file = librsync.PatchedFile(rpout.open("rb"), deltafile)

    if rpnew:
        rpnew.write_from_fileobj(patched_file)
    else:
        write_via_tempfile(patched_file, rpout)



# Copyright 2002 Ben Escoto
#
# This file is part of rdiff-backup.
#
# rdiff-backup is free software; you can redistribute it and/or modify
# under the terms of the GNU General Public License as published by the
# Free Software Foundation; either version 2 of the License, or (at your
# option) any later version.
#
# rdiff-backup is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with rdiff-backup; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA
# 02110-1301, USA
"""Functions to make sure remote requests are kosher"""

import tempfile
from . import Globals, Main, rpath, log


class Violation(Exception):
    """Exception that indicates an improper request has been received"""
    pass


# This will store the list of functions that will be honored from
# remote connections.
allowed_requests = None

# This stores the list of global variables that the client can not
# set on the server.
disallowed_server_globals = ["server", "security_level", "restrict_path"]

# Some common file commands we may want to check to make sure they are
# in the right directory.  Any commands accessing files that could be
# added to allowed_requests must be here.
#
# The keys are files request, the value is the index of the argument
# taking a file.
file_requests = {
    'os.listdir': 0,
    'rpath.make_file_dict': 0,
    'os.chmod': 0,
    'os.chown': 0,
    'os.remove': 0,
    'os.removedirs': 0,
    'os.rename': 0,
    'os.renames': 0,
    'os.rmdir': 0,
    'os.unlink': 0,
    'os.utime': 0,
    'os.lchown': 0,
    'os.link': 1,
    'os.symlink': 1,
    'os.mkdir': 0,
    'os.makedirs': 0,
    'rpath.delete_dir_no_files': 0
}


def initialize(action, cmdpairs):
    """Initialize allowable request list and chroot"""
    global allowed_requests
    set_security_level(action, cmdpairs)
    set_allowed_requests(Globals.security_level)


def reset_restrict_path(rp):
    """Reset restrict path to be within rpath"""
    assert rp.conn is Globals.local_connection
    Globals.restrict_path = rp.normalize().path


def set_security_level(action, cmdpairs):
    """If running client, set security level and restrict_path

    To find these settings, we must look at the action to see what is
    supposed to happen, and then look at the cmdpairs to see what end
    the client is on.

    """

    def islocal(cmdpair):
        return not cmdpair[0]

    def bothlocal(cp1, cp2):
        return islocal(cp1) and islocal(cp2)

    def bothremote(cp1, cp2):
        return not islocal(cp1) and not islocal(cp2)

    def getpath(cmdpair):
        return cmdpair[1]

    if Globals.server:
        return
    cp1 = cmdpairs[0]
    if len(cmdpairs) > 1:
        cp2 = cmdpairs[1]
    else:
        cp2 = cp1

    if action == "backup" or action == "check-destination-dir":
        if bothlocal(cp1, cp2) or bothremote(cp1, cp2):
            sec_level = "minimal"
            rdir = tempfile.gettempdir()
        elif islocal(cp1):
            sec_level = "read-only"
            rdir = getpath(cp1)
        else:
            assert islocal(cp2)
            sec_level = "update-only"
            rdir = getpath(cp2)
    elif action == "restore" or action == "restore-as-of":
        if len(cmdpairs) == 1 or bothlocal(cp1, cp2) or bothremote(cp1, cp2):
            sec_level = "minimal"
            rdir = tempfile.gettempdir()
        elif islocal(cp1):
            sec_level = "read-only"
            Main.restore_set_root(
                rpath.RPath(Globals.local_connection, getpath(cp1)))
            if Main.restore_root:
                rdir = Main.restore_root.path
            else:
                log.Log.FatalError("Invalid restore directory")
        else:
            assert islocal(cp2)
            sec_level = "all"
            rdir = getpath(cp2)
    elif action == "mirror":
        if bothlocal(cp1, cp2) or bothremote(cp1, cp2):
            sec_level = "minimal"
            rdir = tempfile.gettempdir()
        elif islocal(cp1):
            sec_level = "read-only"
            rdir = getpath(cp1)
        else:
            assert islocal(cp2)
            sec_level = "all"
            rdir = getpath(cp2)
    elif action in [
            "test-server", "list-increments", 'list-increment-sizes',
            "list-at-time", "list-changed-since", "calculate-average",
            "remove-older-than", "compare", "compare-hash", "compare-full",
            "verify"
    ]:
        sec_level = "minimal"
        rdir = tempfile.gettempdir()
    else:
        assert 0, "Unknown action %s" % action

    Globals.security_level = sec_level
    Globals.restrict_path = rpath.RPath(Globals.local_connection,
                                        rdir).normalize().path


def set_allowed_requests(sec_level):
    """Set the allowed requests list using the security level"""
    global allowed_requests
    requests = [
        "VirtualFile.readfromid", "VirtualFile.closebyid", "Globals.get",
        "Globals.is_not_None", "Globals.get_dict_val",
        "log.Log.open_logfile_allconn", "log.Log.close_logfile_allconn",
        "Log.log_to_file", "FilenameMapping.set_init_quote_vals_local",
        "FilenameMapping.set_init_quote_vals", "Time.setcurtime_local",
        "SetConnections.add_redirected_conn", "RedirectedRun",
        "sys.stdout.write", "robust.install_signal_handlers"
    ]
    if (sec_level == "read-only" or sec_level == "update-only"
            or sec_level == "all"):
        requests.extend([
            "rpath.make_file_dict", "os.listdir", "rpath.ea_get",
            "rpath.acl_get", "rpath.setdata_local", "log.Log.log_to_file",
            "os.getuid", "rpath.gzip_open_local_read", "rpath.open_local_read",
            "Hardlink.initialize_dictionaries", "user_group.uid2uname",
            "user_group.gid2gname"
        ])
    if sec_level == "read-only" or sec_level == "all":
        requests.extend([
            "fs_abilities.get_readonly_fsa",
            "restore.MirrorStruct.get_increment_times",
            "restore.MirrorStruct.set_mirror_and_rest_times",
            "restore.MirrorStruct.set_mirror_select",
            "restore.MirrorStruct.initialize_rf_cache",
            "restore.MirrorStruct.close_rf_cache",
            "restore.MirrorStruct.get_diffs", "restore.ListChangedSince",
            "restore.ListAtTime", "backup.SourceStruct.get_source_select",
            "backup.SourceStruct.set_source_select",
            "backup.SourceStruct.get_diffs",
            "compare.RepoSide.init_and_get_iter",
            "compare.RepoSide.close_rf_cache", "compare.RepoSide.attach_files",
            "compare.DataSide.get_source_select",
            "compare.DataSide.compare_fast", "compare.DataSide.compare_hash",
            "compare.DataSide.compare_full", "compare.Verify"
        ])
    if sec_level == "update-only" or sec_level == "all":
        requests.extend([
            "log.Log.open_logfile_local", "log.Log.close_logfile_local",
            "log.ErrorLog.open", "log.ErrorLog.isopen", "log.ErrorLog.close",
            "backup.DestinationStruct.set_rorp_cache",
            "backup.DestinationStruct.get_sigs",
            "backup.DestinationStruct.patch_and_increment",
            "Main.backup_touch_curmirror_local",
            "Main.backup_remove_curmirror_local",
            "Main.backup_close_statistics", "regress.check_pids",
            "Globals.ITRB.increment_stat", "statistics.record_error",
            "log.ErrorLog.write_if_open", "fs_abilities.backup_set_globals"
        ])
    if sec_level == "all":
        requests.extend([
            "os.mkdir", "os.chown", "os.lchown", "os.rename", "os.unlink",
            "os.remove", "os.chmod", "os.makedirs",
            "rpath.delete_dir_no_files", "backup.DestinationStruct.patch",
            "restore.TargetStruct.get_initial_iter",
            "restore.TargetStruct.patch",
            "restore.TargetStruct.set_target_select",
            "fs_abilities.restore_set_globals",
            "fs_abilities.single_set_globals", "regress.Regress",
            "manage.delete_earlier_than_local"
        ])
    if Globals.server:
        requests.extend([
            "SetConnections.init_connection_remote", "log.Log.setverbosity",
            "log.Log.setterm_verbosity", "Time.setprevtime_local",
            "Globals.postset_regexp_local",
            "backup.SourceStruct.set_session_info",
            "backup.DestinationStruct.set_session_info",
            "user_group.init_user_mapping", "user_group.init_group_mapping"
        ])
    allowed_requests = {}
    for req in requests:
        allowed_requests[req] = None


def raise_violation(reason, request, arglist):
    """Raise a security violation about given request"""
    raise Violation(
        "\nWARNING: Security Violation!\n"
        "%s for function: %s\n"
        "with arguments: %s\n" % (reason, request.function_string, arglist))


def vet_request(request, arglist):
    """Examine request for security violations"""
    security_level = Globals.security_level
    if security_level == "override":
        return
    if Globals.restrict_path:
        for arg in arglist:
            if isinstance(arg, rpath.RPath):
                _vet_rpath(arg, request, arglist)
        if request.function_string in file_requests:
            vet_filename(request, arglist)
    if request.function_string in allowed_requests:
        return
    if request.function_string in ("Globals.set", "Globals.set_local"):
        if arglist[0] not in disallowed_server_globals:
            return
    raise_violation("Invalid request", request, arglist)


def vet_filename(request, arglist):
    """Check to see if file operation is within the restrict_path"""
    i = file_requests[request.function_string]
    if len(arglist) <= i:
        raise_violation("Argument list shorter than %d" % i + 1, request,
                        arglist)
    filename = arglist[i]
    if not (isinstance(filename, bytes) or isinstance(filename, str)):
        raise_violation("Argument %d doesn't look like a filename" % i,
                        request, arglist)

    _vet_rpath(
        rpath.RPath(Globals.local_connection, filename), request, arglist)


def _vet_rpath(rp, request, arglist):
    """Internal function to validate that a specific path isn't restricted"""
    if Globals.restrict_path and rp.conn is Globals.local_connection:
        normalized, restrict = rp.normalize().path, Globals.restrict_path
        if restrict == b"/":
            return
        components = normalized.split(b"/")
        # 3 cases for restricted dir /usr/foo:  /var, /usr/foobar, /usr/foo/..
        if (not normalized.startswith(restrict)
                or (len(normalized) > len(restrict)
                    and normalized[len(restrict)] != ord("/"))
                or b".." in components):
            raise_violation(
                "Normalized path %s not within restricted path %s" %
                (normalized, restrict), request, arglist)



# Copyright 2002 Ben Escoto
#
# This file is part of rdiff-backup.
#
# rdiff-backup is free software; you can redistribute it and/or modify
# under the terms of the GNU General Public License as published by the
# Free Software Foundation; either version 2 of the License, or (at your
# option) any later version.
#
# rdiff-backup is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with rdiff-backup; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA
# 02110-1301, USA
"""Hold a variety of constants usually set at initialization."""

import re
import os
from . import log

# The current version of rdiff-backup
# Get it from package info or fall back to DEV version.
try:
    import pkg_resources
    version = pkg_resources.get_distribution("rdiff-backup").version
except BaseException:
    version = "DEV"

# If this is set, use this value in seconds as the current time
# instead of reading it from the clock.
current_time = None

# This determines how many bytes to read at a time when copying
blocksize = 131072

# This is used by the BufferedRead class to determine how many
# bytes to request from the underlying file per read().  Larger
# values may save on connection overhead and latency.
conn_bufsize = 393216

# This is used in the CacheCollatedPostProcess and MiscIterToFile
# classes.  The number represents the number of rpaths which may be
# stuck in buffers when moving over a remote connection.
pipeline_max_length = 500

# True if script is running as a server
server = None

# uid and gid of the owner of the rdiff-backup process.  This can
# vary depending on the connection.
try:
    process_uid = os.getuid()
    process_gid = os.getgid()
    process_groups = [process_gid] + os.getgroups()
except AttributeError:
    process_uid = 0
    process_gid = 0
    process_groups = [0]

# If true, when copying attributes, also change target's uid/gid
change_ownership = None

# If true, change the permissions of unwriteable mirror files
# (such as directories) so that they can be written, and then
# change them back.  This defaults to 1 just in case the process
# is not running as root (root doesn't need to change
# permissions).
change_mirror_perms = (process_uid != 0)

# If true, try to reset the atimes of the source partition.
preserve_atime = None

# The following three attributes represent whether extended attributes
# are supported.  If eas_active is true, then the current session
# supports them.  If eas_write is true, then the extended attributes
# should also be written to the destination side.  Finally, eas_conn
# is relative to the current connection, and should be true iff that
# particular connection supports extended attributes.
eas_active = None
eas_write = None
eas_conn = None

# The following settings are like the extended attribute settings, but
# apply to access control lists instead.
acls_active = None
acls_write = None
acls_conn = None

# Like the above, but applies to support of Windows
# access control lists.
win_acls_active = None
win_acls_write = None
win_acls_conn = None

# Like above two setting groups, but applies to support of Mac OS X
# style resource forks.
resource_forks_active = None
resource_forks_write = None
resource_forks_conn = None

# Like the above, but applies to MacOS Carbon Finder creator/type info.
# As of 1.0.2 this has defaulted to off because of bugs
carbonfile_active = None
carbonfile_write = None
carbonfile_conn = None

# This will be set as soon as the LocalConnection class loads
local_connection = None

# All connections should be added to the following list, so
# further global changes can be propagated to the remote systems.
# The first element should be Globals.local_connection.  For a
# server, the second is the connection to the client.
connections = []

# Each process should have a connection number unique to the
# session.  The client has connection number 0.
connection_number = 0

# Dictionary pairing connection numbers with connections.  Set in
# SetConnections for all connections.
connection_dict = {}

# True if the script is the end that reads the source directory
# for backups.  It is true for purely local sessions.
isbackup_reader = None

# Connection of the real backup reader (for which isbackup_reader
# is true)
backup_reader = None

# True if the script is the end that writes to the increment and
# mirror directories.  True for purely local sessions.
isbackup_writer = None

# Connection of the backup writer
backup_writer = None

# Connection of the client
client_conn = None

# When backing up, issource should be true on the reader and isdest on
# the writer.  When restoring, issource should be true on the mirror
# and isdest should be true on the target.
issource = None
isdest = None

# This list is used by the set function below.  When a new
# connection is created with init_connection, its Globals class
# will match this one for all the variables mentioned in this
# list.
changed_settings = []

# The RPath or QuotedRPath of the rdiff-backup-data directory.
rbdir = None

# chars_to_quote is a string whose characters should be quoted.  It
# should be true if certain characters in filenames on the source side
# should be escaped (see FilenameMapping for more info).
chars_to_quote = None
quoting_char = b';'

# evaluate if DOS device names (AUX, PRN, CON, NUL, COM, LPT) should be quoted
# or spaces at the end of file and directory names.
# The default is based on the operating system type (nt or posix).
escape_dos_devices = os.name == 'nt'
escape_trailing_spaces = os.name == 'nt'

# If true, the timestamps use the following format: "2008-09-01T04-49-04-07-00"
# (instead of "2008-09-01T04:49:04-07:00"). This creates timestamps which
# don't need to be escaped on Windows.
use_compatible_timestamps = 0

# Normally there shouldn't be any case of duplicate timestamp but it seems
# we had the issue at some point in time, hence we need the flag to allow
# users to clean up their repository. The default is to abort on such cases.

allow_duplicate_timestamps = False

# If true, emit output intended to be easily readable by a
# computer.  False means output is intended for humans.
parsable_output = None

# If true, then hardlinks will be preserved to mirror and recorded
# in the increments directory.  There is also a difference here
# between None and 0.  When restoring, None or 1 means to preserve
# hardlinks iff can find a hardlink dictionary.  0 means ignore
# hardlink information regardless.
preserve_hardlinks = 1

# If this is false, then rdiff-backup will not compress any
# increments.  Default is to compress based on regexp below.
compression = 1

# Increments based on files whose names match this
# case-insensitive regular expression won't be compressed (applies
# to .snapshots and .diffs).  The second below will be the
# compiled version of the first.
no_compression_regexp_string = (
    b"(?i).*\\.(gz|z|bz|bz2|tgz|zip|zst|rpm|deb|"
    b"jpg|jpeg|gif|png|jp2|mp3|mp4|ogg|ogv|oga|ogm|avi|wmv|mpeg|mpg|rm|mov|mkv|flac|shn|pgp|"
    b"gpg|rz|lz4|lzh|lzo|zoo|lharc|rar|arj|asc|vob|mdf)$")
no_compression_regexp = None

# If true, filelists and directory statistics will be split on
# nulls instead of newlines.
null_separator = None

# Determines whether or not ssh will be run with the -C switch
ssh_compression = 1

# If true, print statistics after successful backup
print_statistics = None

# Controls whether file_statistics file is written in
# rdiff-backup-data dir.  These can sometimes take up a lot of space.
file_statistics = 1

# On the writer connection, the following will be set to the mirror
# Select iterator.
select_mirror = None

# On the backup writer connection, holds the root incrementing branch
# object.  Access is provided to increment error counts.
ITRB = None

# security_level has 4 values and controls which requests from remote
# systems will be honored.  "all" means anything goes. "read-only"
# means that the requests must not write to disk.  "update-only" means
# that requests shouldn't destructively update the disk (but normal
# incremental updates are OK).  "minimal" means only listen to a few
# basic requests.
security_level = "all"

# If this is set, it indicates that the remote connection should only
# deal with paths inside of restrict_path.
restrict_path = None

# If set, a file will be marked as changed if its inode changes.  See
# the man page under --no-compare-inode for more information.
compare_inode = 1

# If set, directories can be fsync'd just like normal files, to
# guarantee that any changes have been comitted to disk.
fsync_directories = None

# If set, exit with error instead of dropping ACLs or ACL entries.
never_drop_acls = None

# Apply this mask to permissions before chmoding.  (Set to 0777 to
# prevent highbit permissions on systems which don't support them.)
permission_mask = 0o7777

# If true, symlinks permissions are affected by the process umask, and
# we should change the umask when creating them in order to preserve
# the original permissions
symlink_perms = None

# If set, the path that should be used instead of the default Python
# tempfile.tempdir value on remote connections
remote_tempdir = None

# Fsync everything by default. Use --no-fsync only if you really know what you are doing
# Not having the data written to disk may render your backup unusable in case of FS failure.
# Using --no-fsync disables only fsync of files during backup and sync() system call upon backup finish
# and pre-regress
do_fsync = True


def get(name):
    """Return the value of something in this module"""
    return globals()[name]


def is_not_None(name):
    """Returns true if value is not None"""
    return globals()[name] is not None


def set(name, val):
    """Set the value of something in this module

    Use this instead of writing the values directly if the setting
    matters to remote sides.  This function updates the
    changed_settings list, so other connections know to copy the
    changes.

    """
    changed_settings.append(name)
    globals()[name] = val


def set_local(name, val):
    """Like set above, but only set current connection"""
    globals()[name] = val


def set_integer(name, val):
    """Like set, but make sure val is an integer"""
    try:
        intval = int(val)
    except ValueError:
        log.Log.FatalError("Variable %s must be set to an integer -\n"
                           "received %s instead." % (name, val))
    set(name, intval)


def set_float(name, val, min=None, max=None, inclusive=1):
    """Like set, but make sure val is float within given bounds"""

    def error():
        s = "Variable %s must be set to a float" % (name, )
        if min is not None and max is not None:
            s += " between %s and %s " % (min, max)
            if inclusive:
                s += "inclusive"
            else:
                s += "not inclusive"
        elif min is not None or max is not None:
            if inclusive:
                inclusive_string = "or equal to "
            else:
                inclusive_string = ""
            if min is not None:
                s += " greater than %s%s" % (inclusive_string, min)
            else:
                s += " less than %s%s" % (inclusive_string, max)
        log.Log.FatalError(s)

    try:
        f = float(val)
    except ValueError:
        error()
    if min is not None:
        if inclusive and f < min:
            error()
        elif not inclusive and f <= min:
            error()
    if max is not None:
        if inclusive and f > max:
            error()
        elif not inclusive and f >= max:
            error()
    set(name, f)


def get_dict_val(name, key):
    """Return val from dictionary in this class"""
    return globals()[name][key]


def set_dict_val(name, key, val):
    """Set value for dictionary in this class"""
    globals()[name][key] = val


def postset_regexp(name, re_string, flags=None):
    """Compile re_string on all existing connections, set to name"""
    for conn in connections:
        conn.Globals.postset_regexp_local(name, re_string, flags)


def postset_regexp_local(name, re_string, flags):
    """Set name to compiled re_string locally"""
    if flags:
        globals()[name] = re.compile(re_string, flags)
    else:
        globals()[name] = re.compile(re_string)



# Copyright 2002 Ben Escoto
#
# This file is part of rdiff-backup.
#
# rdiff-backup is free software; you can redistribute it and/or modify
# under the terms of the GNU General Public License as published by the
# Free Software Foundation; either version 2 of the License, or (at your
# option) any later version.
#
# rdiff-backup is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with rdiff-backup; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA
# 02110-1301, USA
"""Provide time related exceptions and functions"""

import time
import re
import calendar
from . import Globals


class TimeException(Exception):
    pass


_interval_conv_dict = {
    "s": 1,
    "m": 60,
    "h": 3600,
    "D": 86400,
    "W": 7 * 86400,
    "M": 30 * 86400,
    "Y": 365 * 86400
}
_integer_regexp = re.compile("^[0-9]+$")
_session_regexp = re.compile("^[0-9]+B$")
_interval_regexp = re.compile("^([0-9]+)([smhDWMY])")
_genstr_date_regexp1 = re.compile(
    "^(?P<year>[0-9]{4})[-/]"
    "(?P<month>[0-9]{1,2})[-/](?P<day>[0-9]{1,2})$")
_genstr_date_regexp2 = re.compile("^(?P<month>[0-9]{1,2})[-/]"
                                  "(?P<day>[0-9]{1,2})[-/](?P<year>[0-9]{4})$")
curtime = curtimestr = None


def setcurtime(curtime=None):
    """Sets the current time in curtime and curtimestr on all systems"""
    t = curtime or time.time()
    for conn in Globals.connections:
        conn.Time.setcurtime_local(int(t))


def setcurtime_local(timeinseconds):
    """Only set the current time locally"""
    global curtime, curtimestr
    curtime, curtimestr = timeinseconds, timetostring(timeinseconds)


def setprevtime(timeinseconds):
    """Sets the previous inc time in prevtime and prevtimestr"""
    assert 0 < timeinseconds < curtime, \
        "Time %s is out of bounds" % (timeinseconds,)
    timestr = timetostring(timeinseconds)
    for conn in Globals.connections:
        conn.Time.setprevtime_local(timeinseconds, timestr)


def setprevtime_local(timeinseconds, timestr):
    """Like setprevtime but only set the local version"""
    global prevtime, prevtimestr
    prevtime, prevtimestr = timeinseconds, timestr


def timetostring(timeinseconds):
    """Return w3 datetime compliant listing of timeinseconds, or one in
    which :'s have been replaced with -'s"""
    if not Globals.use_compatible_timestamps:
        format_string = "%Y-%m-%dT%H:%M:%S"
    else:
        format_string = "%Y-%m-%dT%H-%M-%S"
    s = time.strftime(format_string, time.localtime(timeinseconds))
    return s + gettzd(timeinseconds)


def timetobytes(timeinseconds):
    return timetostring(timeinseconds).encode('ascii')


def stringtotime(timestring):
    """Return time in seconds from w3 timestring

    If there is an error parsing the string, or it doesn't look
    like a w3 datetime string, return None.

    """

    regexp = re.compile('[-:]')

    try:
        date, daytime = timestring[:19].split("T")
        year, month, day = list(map(int, date.split("-")))
        hour, minute, second = list(map(int, regexp.split(daytime)))
        assert 1900 < year < 2100, year
        assert 1 <= month <= 12
        assert 1 <= day <= 31
        assert 0 <= hour <= 23
        assert 0 <= minute <= 59
        assert 0 <= second <= 61  # leap seconds
        timetuple = (year, month, day, hour, minute, second, -1, -1, 0)
        utc_in_secs = calendar.timegm(timetuple)

        return int(utc_in_secs) + tzdtoseconds(timestring[19:])
    except (TypeError, ValueError, AssertionError):
        return None


def bytestotime(timebytes):
    try:
        return stringtotime(timebytes.decode('ascii'))
    except UnicodeDecodeError:
        return None


def timetopretty(timeinseconds):
    """Return pretty version of time"""
    return time.asctime(time.localtime(timeinseconds))


def stringtopretty(timestring):
    """Return pretty version of time given w3 time string"""
    return timetopretty(stringtotime(timestring))


def prettytotime(prettystring):
    """Converts time like "Mon Jun 5 11:00:23" to epoch sec, or None"""
    try:
        return time.mktime(time.strptime(prettystring))
    except ValueError:
        return None


def inttopretty(seconds):
    """Convert num of seconds to readable string like "2 hours"."""
    partlist = []
    hours, seconds = divmod(seconds, 3600)
    if hours > 1:
        partlist.append("%d hours" % hours)
    elif hours == 1:
        partlist.append("1 hour")

    minutes, seconds = divmod(seconds, 60)
    if minutes > 1:
        partlist.append("%d minutes" % minutes)
    elif minutes == 1:
        partlist.append("1 minute")

    if seconds == 1:
        partlist.append("1 second")
    elif not partlist or seconds > 1:
        if isinstance(seconds, int) or isinstance(seconds, int):
            partlist.append("%s seconds" % seconds)
        else:
            partlist.append("%.2f seconds" % seconds)
    return " ".join(partlist)


def intstringtoseconds(interval_string):
    """Convert a string expressing an interval (e.g. "4D2s") to seconds"""

    def error():
        raise TimeException("""Bad interval string "%s"

Intervals are specified like 2Y (2 years) or 2h30m (2.5 hours).  The
allowed special characters are s, m, h, D, W, M, and Y.  See the man
page for more information.
""" % interval_string)

    if len(interval_string) < 2:
        error()

    total = 0
    while interval_string:
        match = _interval_regexp.match(interval_string)
        if not match:
            error()
        num, ext = int(match.group(1)), match.group(2)
        if ext not in _interval_conv_dict or num < 0:
            error()
        total += num * _interval_conv_dict[ext]
        interval_string = interval_string[match.end(0):]
    return total


def gettzd(timeinseconds=None):
    """Return w3's timezone identification string.

    Expressed as [+/-]hh:mm.  For instance, PDT is -07:00 during
    dayling savings and -08:00 otherwise.  Zone coincides with what
    localtime(), etc., use.  If no argument given, use the current
    time.

    """
    if timeinseconds is None:
        timeinseconds = time.time()
    dst_in_effect = time.daylight and time.localtime(timeinseconds)[8]
    if dst_in_effect:
        offset = -time.altzone / 60
    else:
        offset = -time.timezone / 60
    if offset > 0:
        prefix = "+"
    elif offset < 0:
        prefix = "-"
    else:
        return "Z"  # time is already in UTC

    if Globals.use_compatible_timestamps:
        time_separator = '-'
    else:
        time_separator = ':'
    hours, minutes = list(map(abs, divmod(offset, 60)))
    assert 0 <= hours <= 23
    assert 0 <= minutes <= 59
    return "%s%02d%s%02d" % (prefix, hours, time_separator, minutes)


def tzdtoseconds(tzd):
    """Given w3 compliant TZD, return how far ahead UTC is"""
    if tzd == "Z":
        return 0
    assert len(tzd) == 6  # only accept forms like +08:00 for now
    assert (tzd[0] == "-" or tzd[0] == "+") and (tzd[3] == ":"
                                                 or tzd[3] == "-")
    return -60 * (60 * int(tzd[:3]) + int(tzd[4:]))


def cmp(time1, time2):
    """Compare time1 and time2 and return -1, 0, or 1"""
    if type(time1) is str:
        time1 = stringtotime(time1)
        assert time1 is not None
    if type(time2) is str:
        time2 = stringtotime(time2)
        assert time2 is not None

    if time1 < time2:
        return -1
    elif time1 == time2:
        return 0
    else:
        return 1


def time_from_session(session_num, rp=None):
    """Return time in seconds of given backup

    The current mirror is session_num 0, the next oldest increment has
    number 1, etc.  Requires that the Globals.rbdir directory be set.

    """
    session_times = Globals.rbdir.conn.restore.MirrorStruct \
        .get_increment_times()
    session_times.sort()
    if len(session_times) <= session_num:
        return session_times[0]  # Use oldest if too few backups
    return session_times[-session_num - 1]


def genstrtotime(timestr, curtime=None, rp=None):
    """Convert a generic time string to a time in seconds

    rp is used when the time is of the form "4B" or similar.  Then the
    times of the increments of that particular file are used.

    """
    if curtime is None:
        curtime = globals()['curtime']
    if timestr == "now":
        return curtime

    def error():
        raise TimeException("""Bad time string "%s"

The acceptable time strings are intervals (like "3D64s"), w3-datetime
strings, like "2002-04-26T04:22:01-07:00" (strings like
"2002-04-26T04:22:01" are also acceptable - rdiff-backup will use the
current time zone), or ordinary dates like 2/4/1997 or 2001-04-23
(various combinations are acceptable, but the month always precedes
the day).""" % timestr)

    # Test for straight integer
    if _integer_regexp.search(timestr):
        return int(timestr)

    # Test for w3-datetime format, possibly missing tzd
    t = stringtotime(timestr) or stringtotime(timestr + gettzd())
    if t:
        return t

    # Test for time given as number of backups, like 3B
    if _session_regexp.search(timestr):
        return time_from_session(int(timestr[:-1]), rp)

    # Try for long time, like "Mon Jun 5 11:00:23 1990"
    t = prettytotime(timestr)
    if t is not None:
        return t

    try:  # test for an interval, like "2 days ago"
        return curtime - intstringtoseconds(timestr)
    except TimeException:
        pass

    # Now check for dates like 2001/3/23
    match = _genstr_date_regexp1.search(timestr) or \
        _genstr_date_regexp2.search(timestr)
    if not match:
        error()
    timestr = "%s-%02d-%02dT00:00:00%s" % (match.group('year'),
                                           int(match.group('month')),
                                           int(match.group('day')), gettzd())
    t = stringtotime(timestr)
    if t is not None:
        return t
    else:
        error()



# Copyright 2002 Ben Escoto
#
# This file is part of rdiff-backup.
#
# rdiff-backup is free software; you can redistribute it and/or modify
# under the terms of the GNU General Public License as published by the
# Free Software Foundation; either version 2 of the License, or (at your
# option) any later version.
#
# rdiff-backup is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with rdiff-backup; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA
# 02110-1301, USA
"""Store and retrieve metadata in destination directory

The plan is to store metadata information for all files in the
destination directory in a special metadata file.  There are two
reasons for this:

1)  The filesystem of the mirror directory may not be able to handle
    types of metadata that the source filesystem can.  For instance,
    rdiff-backup may not have root access on the destination side, so
    cannot set uid/gid.  Or the source side may have ACLs and the
    destination side doesn't.

    Hopefully every file system can store binary data.  Storing
    metadata separately allows us to back up anything (ok, maybe
    strange filenames are still a problem).

2)  Metadata can be more quickly read from a file than it can by
    traversing the mirror directory over and over again.  In many
    cases most of rdiff-backup's time is spent comparing metadata (like
    file size and modtime), trying to find differences.  Reading this
    data sequentially from a file is significantly less taxing than
    listing directories and statting files all over the mirror
    directory.

The metadata is stored in a text file, which is a bunch of records
concatenated together.  Each record has the format:

File <filename>
  <field_name1> <value>
  <field_name2> <value>
  ...

Where the lines are separated by newlines.  See the code below for the
field names and values.

"""

import re
import os
import binascii
from . import log, Globals, rpath, Time, rorpiter


class ParsingError(Exception):
    """This is raised when bad or unparsable data is received"""
    pass


def carbonfile2string(cfile):
    """Convert CarbonFile data to a string suitable for storing."""
    if not cfile:
        return "None"
    retvalparts = []
    retvalparts.append('creator:%s' % binascii.hexlify(cfile['creator']))
    retvalparts.append('type:%s' % binascii.hexlify(cfile['type']))
    retvalparts.append('location:%d,%d' % cfile['location'])
    retvalparts.append('flags:%d' % cfile['flags'])
    try:
        retvalparts.append('createDate:%d' % cfile['createDate'])
    except KeyError:
        log.Log("Writing pre-1.1.6 style metadata, without creation date", 9)
    return '|'.join(retvalparts)


def string2carbonfile(data):
    """Re-constitute CarbonFile data from a string stored by
    carbonfile2string."""
    retval = {}
    for component in data.split('|'):
        key, value = component.split(':')
        if key == 'creator':
            retval['creator'] = binascii.unhexlify(value)
        elif key == 'type':
            retval['type'] = binascii.unhexlify(value)
        elif key == 'location':
            a, b = value.split(',')
            retval['location'] = (int(a), int(b))
        elif key == 'flags':
            retval['flags'] = int(value)
        elif key == 'createDate':
            retval['createDate'] = int(value)
    return retval


def RORP2Record(rorpath):
    """From RORPath, return text record of file's metadata"""
    str_list = [b"File %s\n" % quote_path(rorpath.get_indexpath())]

    # Store file type, e.g. "dev", "reg", or "sym", and type-specific data
    type = rorpath.gettype()
    if type is None:
        type = "None"
    str_list.append(b"  Type %b\n" % type.encode('ascii'))
    if type == "reg":
        str_list.append(b"  Size %i\n" % rorpath.getsize())

        # If there is a resource fork, save it.
        if rorpath.has_resource_fork():
            if not rorpath.get_resource_fork():
                rf = b"None"
            else:
                rf = binascii.hexlify(rorpath.get_resource_fork())
            str_list.append(b"  ResourceFork %b\n" % (rf, ))

        # If there is Carbon data, save it.
        if rorpath.has_carbonfile():
            cfile = carbonfile2string(rorpath.get_carbonfile())
            str_list.append(b"  CarbonFile %b\n" % (cfile, ))

        # If file is hardlinked, add that information
        if Globals.preserve_hardlinks != 0:
            numlinks = rorpath.getnumlinks()
            if numlinks > 1:
                str_list.append(b"  NumHardLinks %i\n" % numlinks)
                str_list.append(b"  Inode %i\n" % rorpath.getinode())
                str_list.append(b"  DeviceLoc %i\n" % rorpath.getdevloc())

        # Save any hashes, if available
        if rorpath.has_sha1():
            str_list.append(
                b'  SHA1Digest %b\n' % rorpath.get_sha1().encode('ascii'))

    elif type == "None":
        return b"".join(str_list)
    elif type == "dir" or type == "sock" or type == "fifo":
        pass
    elif type == "sym":
        str_list.append(b"  SymData %b\n" % quote_path(rorpath.readlink()))
    elif type == "dev":
        devchar, major, minor = rorpath.getdevnums()
        str_list.append(
            b"  DeviceNum %b %i %i\n" % (devchar.encode('ascii'), major, minor))

    # Store time information
    if type != 'sym' and type != 'dev':
        str_list.append(b"  ModTime %i\n" % rorpath.getmtime())

    # Add user, group, and permission information
    uid, gid = rorpath.getuidgid()
    str_list.append(b"  Uid %i\n" % uid)
    str_list.append(b"  Uname %b\n" % (rorpath.getuname() or ":").encode())
    str_list.append(b"  Gid %i\n" % gid)
    str_list.append(b"  Gname %b\n" % (rorpath.getgname() or ":").encode())
    str_list.append(b"  Permissions %d\n" % rorpath.getperms())

    # Add long filename information
    if rorpath.has_alt_mirror_name():
        str_list.append(
            b"  AlternateMirrorName %b\n" % (rorpath.get_alt_mirror_name(), ))
    elif rorpath.has_alt_inc_name():
        str_list.append(
            b"  AlternateIncrementName %b\n" % (rorpath.get_alt_inc_name(), ))

    return b"".join(str_list)


line_parsing_regexp = re.compile(b"^ *([A-Za-z0-9]+) (.+)$", re.M)


def Record2RORP(record_string):
    """Given record_string, return RORPath

    For speed reasons, write the RORPath data dictionary directly
    instead of calling rorpath functions.  Profiling has shown this to
    be a time critical function.

    """
    data_dict = {}
    for field, data in line_parsing_regexp.findall(record_string):
        field = field.decode('ascii')
        if field == "File":
            index = quoted_filename_to_index(data)
        elif field == "Type":
            if data == b"None":
                data_dict['type'] = None
            else:
                data_dict['type'] = data.decode('ascii')
        elif field == "Size":
            data_dict['size'] = int(data)
        elif field == "ResourceFork":
            if data == b"None":
                data_dict['resourcefork'] = b""
            else:
                data_dict['resourcefork'] = binascii.unhexlify(data)
        elif field == "CarbonFile":
            if data == b"None":
                data_dict['carbonfile'] = None
            else:
                data_dict['carbonfile'] = string2carbonfile(data)
        elif field == "SHA1Digest":
            data_dict['sha1'] = data.decode('ascii')
        elif field == "NumHardLinks":
            data_dict['nlink'] = int(data)
        elif field == "Inode":
            data_dict['inode'] = int(data)
        elif field == "DeviceLoc":
            data_dict['devloc'] = int(data)
        elif field == "SymData":
            data_dict['linkname'] = unquote_path(data)
        elif field == "DeviceNum":
            devchar, major_str, minor_str = data.split(b" ")
            data_dict['devnums'] = (devchar.decode('ascii'), int(major_str),
                                    int(minor_str))
        elif field == "ModTime":
            data_dict['mtime'] = int(data)
        elif field == "Uid":
            data_dict['uid'] = int(data)
        elif field == "Gid":
            data_dict['gid'] = int(data)
        elif field == "Uname":
            if data == b":" or data == b'None':
                data_dict['uname'] = None
            else:
                data_dict['uname'] = data.decode()
        elif field == "Gname":
            if data == b':' or data == b'None':
                data_dict['gname'] = None
            else:
                data_dict['gname'] = data.decode()
        elif field == "Permissions":
            data_dict['perms'] = int(data)
        elif field == "AlternateMirrorName":
            data_dict['mirrorname'] = data
        elif field == "AlternateIncrementName":
            data_dict['incname'] = data
        else:
            log.Log("Unknown field in line '%s %s'" % (field, data), 2)
    return rpath.RORPath(index, data_dict)


chars_to_quote = re.compile(b"\\n|\\\\")


def quote_path(path_string):
    """Return quoted version of path_string

    Because newlines are used to separate fields in a record, they are
    replaced with \n.  Backslashes become \\ and everything else is
    left the way it is.

    """

    def replacement_func(match_obj):
        """This is called on the match obj of any char that needs quoting"""
        char = match_obj.group(0)
        if char == b"\n":
            return b"\\n"
        elif char == b"\\":
            return b"\\\\"
        assert 0, "Bad char %s needs quoting" % char

    return chars_to_quote.sub(replacement_func, path_string)


def unquote_path(quoted_string):
    """Reverse what was done by quote_path"""

    def replacement_func(match_obj):
        """Unquote match obj of two character sequence"""
        two_chars = match_obj.group(0)
        if two_chars == b"\\n":
            return b"\n"
        elif two_chars == b"\\\\":
            return b"\\"
        log.Log("Warning, unknown quoted sequence %s found" % two_chars, 2)
        return two_chars

    return re.sub(b"\\\\n|\\\\\\\\", replacement_func, quoted_string)


def quoted_filename_to_index(quoted_filename):
    """Return tuple index given quoted filename"""
    if quoted_filename == b'.':
        return ()
    else:
        return tuple(unquote_path(quoted_filename).split(b'/'))


class FlatExtractor:
    """Controls iterating objects from flat file"""

    # Set this in subclass.  record_boundary_regexp should match
    # beginning of next record.  The first group should start at the
    # beginning of the record.  The second group should contain the
    # (possibly quoted) filename.
    record_boundary_regexp = None

    # Set in subclass to function that converts text record to object
    record_to_object = None

    def __init__(self, fileobj):
        self.fileobj = fileobj  # holds file object we are reading from
        self.buf = b""  # holds the next part of the file
        self.at_end = 0  # True if we are at the end of the file
        self.blocksize = 32 * 1024

    def get_next_pos(self):
        """Return position of next record in buffer, or end pos if none"""
        while 1:
            m = self.record_boundary_regexp.search(self.buf, 1)
            if m:
                return m.start(1)
            else:  # add next block to the buffer, loop again
                newbuf = self.fileobj.read(self.blocksize)
                if not newbuf:
                    self.at_end = 1
                    return len(self.buf)
                else:
                    self.buf += newbuf

    def iterate(self):
        """Return iterator that yields all objects with records"""
        for record in self.iterate_records():
            try:
                yield self.record_to_object(record)
            except (ParsingError, ValueError) as e:
                if self.at_end:
                    break  # Ignore whitespace/bad records at end
                log.Log(
                    "Error parsing flat file: %s [%s(%s)]" %
                    (e, type(self), self.fileobj.fileobj.name), 2)

    def iterate_records(self):
        """Yield all text records in order"""
        while 1:
            next_pos = self.get_next_pos()
            if self.at_end:
                if next_pos:
                    yield self.buf[:next_pos]
                break
            yield self.buf[:next_pos]
            self.buf = self.buf[next_pos:]
        assert not self.fileobj.close()

    def skip_to_index(self, index):
        """Scan through the file, set buffer to beginning of index record

        Here we make sure that the buffer always ends in a newline, so
        we will not be splitting lines in half.

        """
        assert not self.buf or self.buf.endswith(b"\n")
        while 1:
            self.buf = self.fileobj.read(self.blocksize)
            self.buf += self.fileobj.readline()
            if not self.buf:
                self.at_end = 1
                return
            while 1:
                m = self.record_boundary_regexp.search(self.buf)
                if not m:
                    break
                cur_index = self.filename_to_index(m.group(2))
                if cur_index >= index:
                    self.buf = self.buf[m.start(1):]
                    return
                else:
                    self.buf = self.buf[m.end(1):]

    def iterate_starting_with(self, index):
        """Iterate objects whose index starts with given index"""
        self.skip_to_index(index)
        if self.at_end:
            return
        while 1:
            next_pos = self.get_next_pos()
            try:
                obj = self.record_to_object(self.buf[:next_pos])
            except (ParsingError, ValueError) as e:
                log.Log("Error parsing metadata file: %s" % (e, ), 2)
            else:
                if obj.index[:len(index)] != index:
                    break
                yield obj
            if self.at_end:
                break
            self.buf = self.buf[next_pos:]
        assert not self.fileobj.close()

    def filename_to_index(self, filename):
        """Translate filename, possibly quoted, into an index tuple

        The filename is the first group matched by
        regexp_boundary_regexp.

        """
        assert 0  # subclass


class RorpExtractor(FlatExtractor):
    """Iterate rorps from metadata file"""
    record_boundary_regexp = re.compile(b"(?:\\n|^)(File (.*?))\\n")
    record_to_object = staticmethod(Record2RORP)
    filename_to_index = staticmethod(quoted_filename_to_index)


class FlatFile:
    """Manage a flat file containing info on various files

    This is used for metadata information, and possibly EAs and ACLs.
    The main read interface is as an iterator.  The storage format is
    a flat, probably compressed file, so random access is not
    recommended.

    Even if the file looks like a text file, it is actually a binary file,
    so that (especially) paths can be stored as bytes, without issue
    with encoding / decoding.
    """
    rp, fileobj, mode = None, None, None
    _buffering_on = 1  # Buffering may be useful because gzip writes are slow
    _record_buffer, _max_buffer_size = None, 100
    _extractor = FlatExtractor  # Override to class that iterates objects
    _object_to_record = None  # Set to function converting object to record
    _prefix = None  # Set to required prefix

    def __init__(self, rp_base, mode, check_path=1, compress=1, callback=None):
        """Open rp (or rp+'.gz') for reading ('r') or writing ('w')

        If callback is available, it will be called on the rp upon
        closing (because the rp may not be known in advance).

        """
        self.mode = mode
        self.callback = callback
        self._record_buffer = []
        if check_path:
            assert (rp_base.isincfile()
                    and rp_base.getincbase_bname() == self._prefix), rp_base
            compress = 1
        if mode == 'r' or mode == 'rb':
            self.rp = rp_base
            self.fileobj = self.rp.open("rb", compress)
        else:
            assert mode == 'w' or mode == 'wb', \
                "File opening mode must be one of r, rb, w or wb, and not %s." % mode
            if compress and check_path and not rp_base.isinccompressed():

                def callback(rp):
                    self.rp = rp

                self.fileobj = rpath.MaybeGzip(rp_base, callback)
            else:
                self.rp = rp_base
                assert not self.rp.lstat(), self.rp
                self.fileobj = self.rp.open("wb", compress=compress)

    def write_record(self, record):
        """Write a (text) record into the file"""
        if self._buffering_on:
            self._record_buffer.append(record)
            if len(self._record_buffer) >= self._max_buffer_size:
                self.fileobj.write(b"".join(self._record_buffer))
                self._record_buffer = []
        else:
            self.fileobj.write(record)

    def write_object(self, object):
        """Convert one object to record and write to file"""
        self.write_record(self._object_to_record(object))

    def get_objects(self, restrict_index=None):
        """Return iterator of objects records from file rp"""
        if not restrict_index:
            return self._extractor(self.fileobj).iterate()
        extractor = self._extractor(self.fileobj)
        return extractor.iterate_starting_with(restrict_index)

    def get_records(self):
        """Return iterator of text records"""
        return self._extractor(self.fileobj).iterate_records()

    def close(self):
        """Close file, for when any writing is done"""
        assert self.fileobj, "File already closed"
        if self._buffering_on and self._record_buffer:
            self.fileobj.write(b"".join(self._record_buffer))
            self._record_buffer = []
        result = self.fileobj.close()
        self.fileobj = None
        self.rp.fsync_with_dir()
        self.rp.setdata()
        if self.callback:
            self.callback(self.rp)
        return result


class MetadataFile(FlatFile):
    """Store/retrieve metadata from mirror_metadata as rorps"""
    _prefix = b"mirror_metadata"
    _extractor = RorpExtractor
    _object_to_record = staticmethod(RORP2Record)


class CombinedWriter:
    """Used for simultaneously writing metadata, eas, and acls"""

    def __init__(self, metawriter, eawriter, aclwriter, winaclwriter):
        self.metawriter = metawriter
        self.eawriter, self.aclwriter, self.winaclwriter = \
            eawriter, aclwriter, winaclwriter  # these can be None

    def write_object(self, rorp):
        """Write information in rorp to all the writers"""
        self.metawriter.write_object(rorp)
        if self.eawriter and not rorp.get_ea().empty():
            self.eawriter.write_object(rorp.get_ea())
        if self.aclwriter and not rorp.get_acl().is_basic():
            self.aclwriter.write_object(rorp.get_acl())
        if self.winaclwriter:
            self.winaclwriter.write_object(rorp.get_win_acl())

    def close(self):
        self.metawriter.close()
        if self.eawriter:
            self.eawriter.close()
        if self.aclwriter:
            self.aclwriter.close()
        if self.winaclwriter:
            self.winaclwriter.close()


class Manager:
    """Read/Combine/Write metadata files by time"""
    meta_prefix = b'mirror_metadata'
    acl_prefix = b'access_control_lists'
    ea_prefix = b'extended_attributes'
    wacl_prefix = b'win_access_control_lists'

    def __init__(self):
        """Set listing of rdiff-backup-data dir"""
        self.rplist = []
        self.timerpmap, self.prefixmap = {}, {}
        for filename in Globals.rbdir.listdir():
            rp = Globals.rbdir.append(filename)
            if rp.isincfile():
                self.add_incrp(rp)

    def add_incrp(self, rp):
        """Add rp to list of inc rps in the rbdir"""
        assert rp.isincfile(), rp
        self.rplist.append(rp)
        time = rp.getinctime()
        if time in self.timerpmap:
            self.timerpmap[time].append(rp)
        else:
            self.timerpmap[time] = [rp]

        incbase = rp.getincbase_bname()
        if incbase in self.prefixmap:
            self.prefixmap[incbase].append(rp)
        else:
            self.prefixmap[incbase] = [rp]

    def _iter_helper(self, prefix, flatfileclass, time, restrict_index):
        """Used below to find the right kind of file by time"""
        if time not in self.timerpmap:
            return None
        for rp in self.timerpmap[time]:
            if rp.getincbase_bname() == prefix:
                return flatfileclass(rp, 'r').get_objects(restrict_index)
        return None

    def get_meta_at_time(self, time, restrict_index):
        """Return iter of metadata rorps at given time (or None)"""
        return self._iter_helper(self.meta_prefix, MetadataFile, time,
                                 restrict_index)

    def get_eas_at_time(self, time, restrict_index):
        """Return Extended Attributes iter at given time (or None)"""
        return self._iter_helper(self.ea_prefix,
                                 eas_acls.ExtendedAttributesFile, time,
                                 restrict_index)

    def get_acls_at_time(self, time, restrict_index):
        """Return ACLs iter at given time from recordfile (or None)"""
        return self._iter_helper(self.acl_prefix,
                                 eas_acls.AccessControlListFile, time,
                                 restrict_index)

    def get_win_acls_at_time(self, time, restrict_index):
        """Return WACLs iter at given time from recordfile (or None)"""
        return self._iter_helper(self.wacl_prefix,
                                 win_acls.WinAccessControlListFile, time,
                                 restrict_index)

    def GetAtTime(self, time, restrict_index=None):
        """Return combined metadata iter with ea/acl info if necessary"""
        cur_iter = self.get_meta_at_time(time, restrict_index)
        if not cur_iter:
            log.Log(
                "Warning, could not find mirror_metadata file.\n"
                "Metadata will be read from filesystem instead.", 2)
            return None

        if Globals.acls_active:
            acl_iter = self.get_acls_at_time(time, restrict_index)
            if not acl_iter:
                log.Log("Warning: Access Control List file not found", 2)
                acl_iter = iter([])
            cur_iter = eas_acls.join_acl_iter(cur_iter, acl_iter)
        if Globals.eas_active:
            ea_iter = self.get_eas_at_time(time, restrict_index)
            if not ea_iter:
                log.Log("Warning: Extended Attributes file not found", 2)
                ea_iter = iter([])
            cur_iter = eas_acls.join_ea_iter(cur_iter, ea_iter)
        if Globals.win_acls_active:
            wacl_iter = self.get_win_acls_at_time(time, restrict_index)
            if not wacl_iter:
                log.Log(
                    "Warning: Windows Access Control List file not"
                    " found.", 2)
                wacl_iter = iter([])
            cur_iter = win_acls.join_wacl_iter(cur_iter, wacl_iter)

        return cur_iter

    def _writer_helper(self, prefix, flatfileclass, typestr, time):
        """Used in the get_xx_writer functions, returns a writer class"""
        if time is None:
            timestr = Time.curtimestr
        else:
            timestr = Time.timetobytes(time)
        triple = map(os.fsencode, (prefix, timestr, typestr))
        filename = b'.'.join(triple)
        rp = Globals.rbdir.append(filename)
        assert not rp.lstat(), "File %s already exists!" % (rp.path, )
        assert rp.isincfile()
        return flatfileclass(rp, 'w', callback=self.add_incrp)

    def get_meta_writer(self, typestr, time):
        """Return MetadataFile object opened for writing at given time"""
        return self._writer_helper(self.meta_prefix, MetadataFile, typestr,
                                   time)

    def get_ea_writer(self, typestr, time):
        """Return ExtendedAttributesFile opened for writing"""
        return self._writer_helper(
            self.ea_prefix, eas_acls.ExtendedAttributesFile, typestr, time)

    def get_acl_writer(self, typestr, time):
        """Return AccessControlListFile opened for writing"""
        return self._writer_helper(
            self.acl_prefix, eas_acls.AccessControlListFile, typestr, time)

    def get_win_acl_writer(self, typestr, time):
        """Return WinAccessControlListFile opened for writing"""
        return self._writer_helper(
            self.wacl_prefix, win_acls.WinAccessControlListFile, typestr, time)

    def GetWriter(self, typestr=b'snapshot', time=None):
        """Get a writer object that can write meta and possibly acls/eas"""
        metawriter = self.get_meta_writer(typestr, time)
        if not Globals.eas_active and not Globals.acls_active and \
           not Globals.win_acls_active:
            return metawriter  # no need for a CombinedWriter

        if Globals.eas_active:
            ea_writer = self.get_ea_writer(typestr, time)
        else:
            ea_writer = None
        if Globals.acls_active:
            acl_writer = self.get_acl_writer(typestr, time)
        else:
            acl_writer = None
        if Globals.win_acls_active:
            win_acl_writer = self.get_win_acl_writer(typestr, time)
        else:
            win_acl_writer = None
        return CombinedWriter(metawriter, ea_writer, acl_writer,
                              win_acl_writer)


class PatchDiffMan(Manager):
    """Contains functions for patching and diffing metadata

    To save space, we can record a full list of only the most recent
    metadata, using the normal rdiff-backup reverse increment
    strategy.  Instead of using librsync to compute diffs, though, we
    use our own technique so that the diff files are still
    hand-editable.

    A mirror_metadata diff has the same format as a mirror_metadata
    snapshot.  If the record for an index is missing from the diff, it
    indicates no change from the original.  If it is present it
    replaces the mirror_metadata entry, unless it has Type None, which
    indicates the record should be deleted from the original.

    """
    max_diff_chain = 9  # After this many diffs, make a new snapshot

    def get_diffiter(self, new_iter, old_iter):
        """Iterate meta diffs of new_iter -> old_iter"""
        for new_rorp, old_rorp in rorpiter.Collate2Iters(new_iter, old_iter):
            if not old_rorp:
                yield rpath.RORPath(new_rorp.index)
            elif not new_rorp or new_rorp.data != old_rorp.data:
                # exact compare here, can't use == on rorps
                yield old_rorp

    def sorted_prefix_inclist(self, prefix, min_time=0):
        """Return reverse sorted (by time) list of incs with given prefix"""
        if prefix not in self.prefixmap:
            return []
        sortlist = [(rp.getinctime(), rp) for rp in self.prefixmap[prefix]]

        # we sort before we validate against duplicates so that we tell
        # first about the youngest case of duplication
        sortlist.sort(reverse=True, key=lambda x: x[0])

        # we had cases where the timestamp of the metadata files were
        # duplicates, we need to fail or at least warn about such cases
        unique_set = set()
        for (time, rp) in sortlist:
            if time in unique_set:
                if Globals.allow_duplicate_timestamps:
                    log.Log("Warning: metadata file '%s' has a duplicate "
                            "timestamp date, you might not be able to "
                            "recover files on or earlier than this date. "
                            "Assuming you're in the process of cleaning up "
                            "your repository." %
                            rp.get_safepath(), 2)
                else:
                    log.Log.FatalError(
                        "Metadata file '%s' has a duplicate timestamp date, "
                        "you might not be able to recover files on or earlier "
                        "than this date. "
                        "Check the man page on how to clean up your repository "
                        "using the '--allow-duplicate-timestamps' option." %
                        rp.get_safepath())
            else:
                unique_set.add(time)

        return [rp for (time, rp) in sortlist if time >= min_time]

    def check_needs_diff(self):
        """Check if we should diff, returns (new, old) rps, or (None, None)"""
        inclist = self.sorted_prefix_inclist(b'mirror_metadata')
        assert len(inclist) >= 1
        if len(inclist) == 1:
            return (None, None)
        newrp, oldrp = inclist[:2]
        assert newrp.getinctype() == oldrp.getinctype() == b'snapshot'

        chainlen = 1
        for rp in inclist[2:]:
            if rp.getinctype() != b'diff':
                break
            chainlen += 1
        if chainlen >= self.max_diff_chain:
            return (None, None)
        return (newrp, oldrp)

    def ConvertMetaToDiff(self):
        """Replace a mirror snapshot with a diff if it's appropriate"""
        newrp, oldrp = self.check_needs_diff()
        if not newrp:
            return
        log.Log("Writing mirror_metadata diff", 6)

        diff_writer = self.get_meta_writer(b'diff', oldrp.getinctime())
        new_iter = MetadataFile(newrp, 'r').get_objects()
        old_iter = MetadataFile(oldrp, 'r').get_objects()
        for diff_rorp in self.get_diffiter(new_iter, old_iter):
            diff_writer.write_object(diff_rorp)
        diff_writer.close()  # includes sync
        oldrp.delete()

    def get_meta_at_time(self, time, restrict_index):
        """Get metadata rorp iter, possibly by patching with diffs"""
        meta_iters = [
            MetadataFile(rp, 'r').get_objects(restrict_index)
            for rp in self.relevant_meta_incs(time)
        ]
        if not meta_iters:
            return None
        if len(meta_iters) == 1:
            return meta_iters[0]
        return self.iterate_patched_meta(meta_iters)

    def relevant_meta_incs(self, time):
        """Return list [snapshotrp, diffrps ...] time sorted"""
        inclist = self.sorted_prefix_inclist(b'mirror_metadata', min_time=time)
        if not inclist:
            return inclist
        assert inclist[-1].getinctime() == time, inclist[-1]
        for i in range(len(inclist) - 1, -1, -1):
            if inclist[i].getinctype() == b'snapshot':
                return inclist[i:]
        assert 0, "Inclist %s contains no snapshots" % (inclist, )

    def iterate_patched_meta(self, meta_iter_list):
        """Return an iter of metadata rorps by combining the given iters

        The iters should be given as a list/tuple in reverse
        chronological order.  The earliest rorp in each iter will
        supercede all the later ones.

        """
        for meta_tuple in rorpiter.CollateIterators(*meta_iter_list):
            for i in range(len(meta_tuple) - 1, -1, -1):
                if meta_tuple[i]:
                    if meta_tuple[i].lstat():
                        yield meta_tuple[i]
                    break  # move to next index
            else:
                assert 0, "No valid rorps"


ManagerObj = None  # Set this later to Manager instance


def SetManager():
    global ManagerObj
    ManagerObj = PatchDiffMan()
    return ManagerObj


from . import eas_acls, win_acls  # noqa: E402



# Copyright 2002 Ben Escoto
#
# This file is part of rdiff-backup.
#
# rdiff-backup is free software; you can redistribute it and/or modify
# under the terms of the GNU General Public License as published by the
# Free Software Foundation; either version 2 of the License, or (at your
# option) any later version.
#
# rdiff-backup is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with rdiff-backup; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA
# 02110-1301, USA
"""list, delete, and otherwise manage increments"""

import os
from .log import Log
from . import Globals, Time, statistics, restore, selection, FilenameMapping


class ManageException(Exception):
    pass


def get_file_type(rp):
    """Returns one of "regular", "directory", "missing", or "special"."""
    if not rp.lstat():
        return "missing"
    elif rp.isdir():
        return "directory"
    elif rp.isreg():
        return "regular"
    else:
        return "special"


def get_inc_type(inc):
    """Return file type increment represents"""
    assert inc.isincfile()
    type = inc.getinctype()
    if type == b"dir":
        return "directory"
    elif type == b"diff":
        return "regular"
    elif type == b"missing":
        return "missing"
    elif type == b"snapshot":
        return get_file_type(inc)
    else:
        assert None, "Unknown type %s" % type


def describe_incs_parsable(incs, mirror_time, mirrorrp):
    """Return a string parsable by computer describing the increments

    Each line is a time in seconds of the increment, and then the
    type of the file.  It will be sorted oldest to newest.  For example:

    10000 regular
    20000 directory
    30000 special
    40000 missing
    50000 regular    <- last will be the current mirror

    """
    incpairs = [(inc.getinctime(), inc) for inc in incs]
    incpairs.sort()
    result = ["%s %s" % (time, get_inc_type(inc)) for time, inc in incpairs]
    result.append("%s %s" % (mirror_time, get_file_type(mirrorrp)))
    return "\n".join(result)


def describe_incs_human(incs, mirror_time, mirrorrp):
    """Return a string describing all the the root increments"""
    incpairs = [(inc.getinctime(), inc) for inc in incs]
    incpairs.sort()

    result = ["Found %d increments:" % len(incpairs)]
    if Globals.chars_to_quote:
        for time, inc in incpairs:
            result.append("    %s   %s" % (
                os.fsdecode(FilenameMapping.unquote(inc.dirsplit()[1])),
                Time.timetopretty(time)))
    else:
        for time, inc in incpairs:
            result.append("    %s   %s" % (
                os.fsdecode(inc.dirsplit()[1]),
                Time.timetopretty(time)))
    result.append("Current mirror: %s" % Time.timetopretty(mirror_time))
    return "\n".join(result)


def delete_earlier_than(baserp, time):
    """Deleting increments older than time in directory baserp

    time is in seconds.  It will then delete any empty directories
    in the tree.  To process the entire backup area, the
    rdiff-backup-data directory should be the root of the tree.

    """
    baserp.conn.manage.delete_earlier_than_local(baserp, time)


def delete_earlier_than_local(baserp, time):
    """Like delete_earlier_than, but run on local connection for speed"""
    assert baserp.conn is Globals.local_connection

    def yield_files(rp):
        if rp.isdir():
            for filename in rp.listdir():
                for sub_rp in yield_files(rp.append(filename)):
                    yield sub_rp
        yield rp

    for rp in yield_files(baserp):
        if ((rp.isincfile() and rp.getinctime() < time)
                or (rp.isdir() and not rp.listdir())):
            Log("Deleting increment file %s" % rp.get_safepath(), 5)
            rp.delete()


class IncObj:
    """Increment object - represent a completed increment"""

    def __init__(self, incrp):
        """IncObj initializer

        incrp is an RPath of a path like increments.TIMESTR.dir
        standing for the root of the increment.

        """
        if not incrp.isincfile():
            raise ManageException(
                "%s is not an inc file" % incrp.get_safepath())
        self.incrp = incrp
        self.time = incrp.getinctime()

    def getbaserp(self):
        """Return rp of the incrp without extensions"""
        return self.incrp.getincbase()

    def pretty_time(self):
        """Return a formatted version of inc's time"""
        return Time.timetopretty(self.time)


def ListIncrementSizes(mirror_root, index):
    """Return string summarizing the size of all the increments"""
    stat_obj = statistics.StatsObj()  # used for byte summary string

    def get_total(rp_iter):
        """Return the total size of everything in rp_iter"""
        total = 0
        for rp in rp_iter:
            total += rp.getsize()
        return total

    def get_time_dict(inc_iter):
        """Return dictionary pairing times to total size of incs"""
        time_dict = {}
        for inc in inc_iter:
            if not inc.isincfile():
                continue
            t = inc.getinctime()
            if t not in time_dict:
                time_dict[t] = 0
            time_dict[t] += inc.getsize()
        return time_dict

    def get_mirror_select():
        """Return iterator of mirror rpaths"""
        mirror_base = mirror_root.new_index(index)
        mirror_select = selection.Select(mirror_base)
        if not index:  # must exclude rdiff-backup-directory
            mirror_select.parse_rbdir_exclude()
        return mirror_select.set_iter()

    def get_inc_select():
        """Return iterator of increment rpaths"""
        inc_base = Globals.rbdir.append_path(b'increments', index)
        for base_inc in restore.get_inclist(inc_base):
            yield base_inc
        if inc_base.isdir():
            inc_select = selection.Select(inc_base).set_iter()
            for inc in inc_select:
                yield inc

    def get_summary_triples(mirror_total, time_dict):
        """Return list of triples (time, size, cumulative size)"""
        triples = []

        cur_mir_base = Globals.rbdir.append(b'current_mirror')
        mirror_time = restore.get_inclist(cur_mir_base)[0].getinctime()
        triples.append((mirror_time, mirror_total, mirror_total))

        inc_times = list(time_dict.keys())
        inc_times.sort()
        inc_times.reverse()
        cumulative_size = mirror_total
        for inc_time in inc_times:
            size = time_dict[inc_time]
            cumulative_size += size
            triples.append((inc_time, size, cumulative_size))
        return triples

    def triple_to_line(triple):
        """Convert triple to display string"""
        time, size, cum_size = triple
        return "%24s   %13s   %15s" % \
            (Time.timetopretty(time),
             stat_obj.get_byte_summary_string(size),
             stat_obj.get_byte_summary_string(cum_size))

    mirror_total = get_total(get_mirror_select())
    time_dict = get_time_dict(get_inc_select())
    triples = get_summary_triples(mirror_total, time_dict)

    sizes = [
        '%12s %9s  %15s   %20s' % ('Time', '', 'Size', 'Cumulative size'),
        '-' * 77,
        triple_to_line(triples[0]) + '   (current mirror)'
    ]
    for triple in triples[1:]:
        sizes.append(triple_to_line(triple))
    return '\n'.join(sizes)



# Copyright 2002, 2003 Ben Escoto
#
# This file is part of rdiff-backup.
#
# rdiff-backup is free software; you can redistribute it and/or modify
# under the terms of the GNU General Public License as published by the
# Free Software Foundation; either version 2 of the License, or (at your
# option) any later version.
#
# rdiff-backup is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with rdiff-backup; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA
# 02110-1301, USA
"""Parse args and setup connections

The functions in this module are used once by Main to parse file
descriptions like bescoto@folly.stanford.edu:/usr/bin/ls and to set up
the related connections.

"""

import os
import re
import sys
import subprocess
from .log import Log
from . import Globals, connection, rpath

# This is the schema that determines how rdiff-backup will open a
# pipe to the remote system.  If the file is given as A::B, %s will
# be substituted with A in the schema.
__cmd_schema = b'ssh -C %s rdiff-backup --server'
__cmd_schema_no_compress = b'ssh %s rdiff-backup --server'

# This is a list of remote commands used to start the connections.
# The first is None because it is the local connection.
__conn_remote_cmds = [None]


class SetConnectionsException(Exception):
    pass


def get_cmd_pairs(arglist, remote_schema=None, remote_cmd=None):
    """Map the given file descriptions into command pairs

    Command pairs are tuples cmdpair with length 2.  cmdpair[0] is
    None iff it describes a local path, and cmdpair[1] is the path.

    """
    global __cmd_schema
    if remote_schema:
        __cmd_schema = remote_schema
    elif not Globals.ssh_compression:
        __cmd_schema = __cmd_schema_no_compress

    if Globals.remote_tempdir:
        __cmd_schema += (b" --tempdir=" + Globals.remote_tempdir)

    if not arglist:
        return []
    desc_pairs = list(map(parse_file_desc, arglist))

    if [x for x in desc_pairs if x[0]]:  # True if any host_info found
        if remote_cmd:
            Log.FatalError("The --remote-cmd flag is not compatible "
                           "with remote file descriptions.")
    elif remote_schema:
        Log("Remote schema option ignored - no remote file "
            "descriptions.", 2)
    cmd_pairs = list(map(desc2cmd_pairs, desc_pairs))
    if remote_cmd:  # last file description gets remote_cmd
        cmd_pairs[-1] = (remote_cmd, cmd_pairs[-1][1])
    return cmd_pairs


def cmdpair2rp(cmd_pair):
    """Return normalized RPath from cmd_pair (remote_cmd, filename)"""
    cmd, filename = cmd_pair
    if cmd:
        conn = init_connection(cmd)
    else:
        conn = Globals.local_connection
    return rpath.RPath(conn, filename).normalize()


def desc2cmd_pairs(desc_pair):
    """Return pair (remote_cmd, filename) from desc_pair"""
    host_info, filename = desc_pair
    if not host_info:
        return (None, filename)
    else:
        return (fill_schema(host_info), filename)


def parse_file_desc(file_desc):
    """Parse file description returning pair (host_info, filename)

    In other words, bescoto@folly.stanford.edu::/usr/bin/ls =>
    ("bescoto@folly.stanford.edu", "/usr/bin/ls").  The
    complication is to allow for quoting of : by a \\.  If the
    string is not separated by :, then the host_info is None.

    """

    # paths and similar objects must always be bytes
    file_desc = os.fsencode(file_desc)
    # match double colon not preceded by an odd number of backslashes
    file_match = re.fullmatch(rb"^(?P<host>.*[^\\](?:\\\\)*)::(?P<path>.*)$",
                              file_desc)
    if file_match:
        file_host = file_match.group("host")
        # According to description, the backslashes must be unquoted, i.e.
        # double backslashes replaced by single ones, and single ones removed.
        # Hence we split along double ones, remove single ones in each element,
        # and join back with a single backslash.
        file_host = b'\\'.join(
            [x.replace(b'\\', b'') for x in re.split(rb'\\\\', file_host) if x])
        file_path = file_match.group("path")
    else:
        if re.match(rb"^::", file_desc):
            raise SetConnectionsException("No file host in '%s'" % file_desc)
        file_host = None
        file_path = file_desc

    # make sure paths under Windows use / instead of \
    if os.path.altsep:  # only Windows has an alternative separator for paths
        file_path = file_path.replace(os.fsencode(os.path.sep), b'/')

    if not file_path:
        raise SetConnectionsException("No file path in '%s'" % file_desc)

    return (file_host, file_path)


def fill_schema(host_info):
    """Fills host_info into the schema and returns remote command"""
    try:
        return __cmd_schema % host_info
    except TypeError:
        Log.FatalError("Invalid remote schema:\n\n%s\n" % _safe_str(__cmd_schema))


def init_connection(remote_cmd):
    """Run remote_cmd, register connection, and then return it

    If remote_cmd is None, then the local connection will be
    returned.  This also updates some settings on the remote side,
    like global settings, its connection number, and verbosity.

    """
    if not remote_cmd:
        return Globals.local_connection

    Log("Executing %s" % _safe_str(remote_cmd), 4)
    try:
        # we need buffered read on SSH communications, hence using
        # default value for bufsize parameter
        if os.name == 'nt':
            # FIXME workaround because python 3.7 doesn't yet accept bytes
            process = subprocess.Popen(
                os.fsdecode(remote_cmd),
                shell=True,
                stdin=subprocess.PIPE,
                stdout=subprocess.PIPE)
        else:
            process = subprocess.Popen(
                remote_cmd,
                shell=True,
                stdin=subprocess.PIPE,
                stdout=subprocess.PIPE)
        (stdin, stdout) = (process.stdin, process.stdout)
    except OSError:
        (stdin, stdout) = (None, None)
    conn_number = len(Globals.connections)
    conn = connection.PipeConnection(stdout, stdin, conn_number)

    check_connection_version(conn, remote_cmd)
    Log("Registering connection %d" % conn_number, 7)
    init_connection_routing(conn, conn_number, remote_cmd)
    init_connection_settings(conn)
    return conn


def check_connection_version(conn, remote_cmd):
    """Log warning if connection has different version"""
    try:
        remote_version = conn.Globals.get('version')
    except connection.ConnectionError as exception:
        Log.FatalError("""%s

Couldn't start up the remote connection by executing

    %s

Remember that, under the default settings, rdiff-backup must be
installed in the PATH on the remote system.  See the man page for more
information on this.  This message may also be displayed if the remote
version of rdiff-backup is quite different from the local version (%s).""" %
                       (exception, _safe_str(remote_cmd), Globals.version))
    except OverflowError:
        Log.FatalError(
            """Integer overflow while attempting to establish the
remote connection by executing

    %s

Please make sure that nothing is printed (e.g., by your login shell) when this
command executes. Try running this command:

    %s

which should only print out the text: rdiff-backup <version>""" %
            (_safe_str(remote_cmd),
             _safe_str(remote_cmd.replace(b"--server", b"--version"))))

    if remote_version != Globals.version:
        Log(
            "Warning: Local version %s does not match remote version %s." %
            (Globals.version, remote_version), 2)


def init_connection_routing(conn, conn_number, remote_cmd):
    """Called by init_connection, establish routing, conn dict"""
    Globals.connection_dict[conn_number] = conn

    conn.SetConnections.init_connection_remote(conn_number)
    for other_remote_conn in Globals.connections[1:]:
        conn.SetConnections.add_redirected_conn(other_remote_conn.conn_number)
        other_remote_conn.SetConnections.add_redirected_conn(conn_number)

    Globals.connections.append(conn)
    __conn_remote_cmds.append(remote_cmd)


def init_connection_settings(conn):
    """Tell new conn about log settings and updated globals"""
    conn.log.Log.setverbosity(Log.verbosity)
    conn.log.Log.setterm_verbosity(Log.term_verbosity)
    for setting_name in Globals.changed_settings:
        conn.Globals.set(setting_name, Globals.get(setting_name))


def init_connection_remote(conn_number):
    """Run on server side to tell self that have given conn_number"""
    Globals.connection_number = conn_number
    Globals.local_connection.conn_number = conn_number
    Globals.connection_dict[0] = Globals.connections[1]
    Globals.connection_dict[conn_number] = Globals.local_connection


def add_redirected_conn(conn_number):
    """Run on server side - tell about redirected connection"""
    Globals.connection_dict[conn_number] = \
        connection.RedirectedConnection(conn_number)


def UpdateGlobal(setting_name, val):
    """Update value of global variable across all connections"""
    for conn in Globals.connections:
        conn.Globals.set(setting_name, val)


def BackupInitConnections(reading_conn, writing_conn):
    """Backup specific connection initialization"""
    reading_conn.Globals.set("isbackup_reader", 1)
    writing_conn.Globals.set("isbackup_writer", 1)
    UpdateGlobal("backup_reader", reading_conn)
    UpdateGlobal("backup_writer", writing_conn)


def CloseConnections():
    """Close all connections.  Run by client"""
    assert not Globals.server
    for conn in Globals.connections:
        conn.quit()
    del Globals.connections[1:]  # Only leave local connection
    Globals.connection_dict = {0: Globals.local_connection}
    Globals.backup_reader = Globals.isbackup_reader = \
        Globals.backup_writer = Globals.isbackup_writer = None


def TestConnections(rpaths):
    """Test connections, printing results"""
    if len(Globals.connections) == 1:
        print("No remote connections specified")
    else:
        assert len(Globals.connections) == len(rpaths) + 1, \
            "All %d parameters must be remote of the form 'server::path'." % \
            len(rpaths)
        for i in range(1, len(Globals.connections)):
            test_connection(i, rpaths[i - 1])


def test_connection(conn_number, rp):
    """Test connection.  conn_number 0 is the local connection"""
    print("Testing server started by: ", __conn_remote_cmds[conn_number])
    conn = Globals.connections[conn_number]
    try:
        assert conn.Globals.get('current_time') is None
        try:
            assert type(conn.os.getuid()) is int
        except AttributeError:  # Windows doesn't support os.getuid()
            assert type(conn.os.listdir(rp.path)) is list
        version = conn.Globals.get('version')
    except BaseException:
        sys.stderr.write("Server tests failed\n")
        raise
    if not version == Globals.version:
        print("""Server may work, but there is a version mismatch:
Local version: %s
Remote version: %s

In general, an attempt is made to guarantee compatibility only between
different minor versions of the same stable series.  For instance, you
should expect 0.12.4 and 0.12.7 to be compatible, but not 0.12.7
and 0.13.3, nor 0.13.2 and 0.13.4.
""" % (Globals.version, version))
    else:
        print("Server OK")


def _safe_str(cmd):
    """Transform bytes into string without risk of conversion error"""
    if isinstance(cmd, str):
        return cmd
    else:
        return str(cmd, errors='replace')



# Copyright 2002 Ben Escoto
#
# This file is part of rdiff-backup.
#
# rdiff-backup is free software; you can redistribute it and/or modify
# under the terms of the GNU General Public License as published by the
# Free Software Foundation; either version 2 of the License, or (at your
# option) any later version.
#
# rdiff-backup is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with rdiff-backup; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA
# 02110-1301, USA
"""Manage logging, displaying and recording messages with required verbosity"""

import datetime
import sys
import traceback
import types
import re
import os  # needed to grab verbosity as environment variable
from . import Globals, rpath

LOGFILE_ENCODING = 'utf-8'


def _to_bytes(logline, encoding=LOGFILE_ENCODING):
    """
    Convert string into bytes for logging into file.
    """
    assert logline
    assert isinstance(logline, str)
    return logline.encode(encoding, 'backslashreplace')


class LoggerError(Exception):
    pass


class Logger:
    """All functions which deal with logging"""

    def __init__(self):
        self.log_file_open = None
        self.log_file_local = None
        self.verbosity = self.term_verbosity = int(
            os.getenv('RDIFF_BACKUP_VERBOSITY', '3'))
        # termverbset is true if the term_verbosity has been explicitly set
        self.termverbset = None

    def setverbosity(self, verbosity_string):
        """Set verbosity levels.  Takes a number string"""
        try:
            self.verbosity = int(verbosity_string)
        except ValueError:
            Log.FatalError("Verbosity must be a number, received '%s' "
                           "instead." % verbosity_string)
        if not self.termverbset:
            self.term_verbosity = self.verbosity

    def setterm_verbosity(self, termverb_string):
        """Set verbosity to terminal.  Takes a number string"""
        try:
            self.term_verbosity = int(termverb_string)
        except ValueError:
            Log.FatalError("Terminal verbosity must be a number, received "
                           "'%s' instead." % termverb_string)
        self.termverbset = 1

    def open_logfile(self, rpath):
        """Inform all connections of an open logfile.

        rpath.conn will write to the file, and the others will pass
        write commands off to it.

        """
        assert not self.log_file_open
        rpath.conn.log.Log.open_logfile_local(rpath)
        for conn in Globals.connections:
            conn.log.Log.open_logfile_allconn(rpath.conn)

    def open_logfile_allconn(self, log_file_conn):
        """Run on all connections to signal log file is open"""
        self.log_file_open = 1
        self.log_file_conn = log_file_conn

    def open_logfile_local(self, rpath):
        """Open logfile locally - should only be run on one connection"""
        assert rpath.conn is Globals.local_connection
        try:
            self.logfp = rpath.open("ab")
        except (OSError, IOError) as e:
            raise LoggerError(
                "Unable to open logfile %s: %s" % (rpath.path, e))
        self.log_file_local = 1
        self.logrp = rpath

    def close_logfile(self):
        """Close logfile and inform all connections"""
        if self.log_file_open:
            for conn in Globals.connections:
                conn.log.Log.close_logfile_allconn()
            self.log_file_conn.log.Log.close_logfile_local()

    def close_logfile_allconn(self):
        """Run on every connection"""
        self.log_file_open = None

    def close_logfile_local(self):
        """Run by logging connection - close logfile"""
        assert self.log_file_conn is Globals.local_connection
        assert not self.logfp.close()
        self.log_file_local = None

    def format(self, message, verbosity):
        """Format the message, possibly adding date information"""
        if verbosity < 9:
            return "%s\n" % message
        else:
            timestamp = datetime.datetime.now(
                datetime.timezone.utc).astimezone().strftime(
                    "%F %H:%M:%S.%f %z")
            if Globals.server:
                role = "SERVER"
            else:
                role = "CLIENT"
            return "%s  <%s-%d>  %s\n" % (timestamp, role, os.getpid(), message)

    def __call__(self, message, verbosity):
        """Log message that has verbosity importance

        message can be a string, which is logged as-is, or a function,
        which is then called and should return the string to be
        logged.  We do it this way in case producing the string would
        take a significant amount of CPU.

        """
        if verbosity > self.verbosity and verbosity > self.term_verbosity:
            return

        if not isinstance(message, (bytes, str)):
            assert isinstance(message, types.FunctionType)
            message = message()

        if verbosity <= self.verbosity:
            self.log_to_file(message)
        if verbosity <= self.term_verbosity:
            self.log_to_term(message, verbosity)

    def log_to_file(self, message):
        """Write the message to the log file, if possible"""
        if self.log_file_open:
            if self.log_file_local:
                tmpstr = self.format(message, self.verbosity)
                self.logfp.write(_to_bytes(tmpstr))
                self.logfp.flush()
            else:
                self.log_file_conn.log.Log.log_to_file(message)

    def log_to_term(self, message, verbosity):
        """Write message to stdout/stderr"""
        if verbosity <= 2 or Globals.server:
            termfp = sys.stderr.buffer
        else:
            termfp = sys.stdout.buffer
        tmpstr = self.format(message, self.term_verbosity)
        termfp.write(_to_bytes(tmpstr, encoding=sys.stdout.encoding))

    def conn(self, direction, result, req_num):
        """Log some data on the connection

        The main worry with this function is that something in here
        will create more network traffic, which will spiral to
        infinite regress.  So, for instance, logging must only be done
        to the terminal, because otherwise the log file may be remote.

        """
        if self.term_verbosity < 9:
            return
        if type(result) is bytes:
            result_repr = repr(result)
        else:
            result_repr = str(result)
        # shorten the result to a max size of 720 chars with ellipsis if needed
        #result_repr = result_repr[:720] + (result_repr[720:] and '[...]')  # noqa: E265
        if Globals.server:
            conn_str = "Server"
        else:
            conn_str = "Client"
        self.log_to_term(
            "%s %s (%d): %s" % (conn_str, direction, req_num, result_repr), 9)

    def FatalError(self, message, no_fatal_message=0, errlevel=1):
        """Log a fatal error and exit"""
        assert no_fatal_message == 0 or no_fatal_message == 1
        if no_fatal_message:
            prefix_string = ""
        else:
            prefix_string = "Fatal Error: "
        self.log_to_term(prefix_string + message, 1)
        sys.exit(errlevel)

    def exception_to_string(self, arglist=[]):
        """Return string version of current exception plus what's in arglist"""
        type, value, tb = sys.exc_info()
        s = ("Exception '%s' raised of class '%s':\n%s" %
             (value, type, "".join(traceback.format_tb(tb))))
        if arglist:
            s += "__Arguments:"
            for arg in arglist:
                s += "\n"
                try:
                    s += str(arg)
                except UnicodeError:
                    s += str(arg).encode('ascii', 'replace')
        return s

    def exception(self, only_terminal=0, verbosity=5):
        """Log an exception and traceback

        If only_terminal is None, log normally.  If it is 1, then only
        log to disk if log file is local (self.log_file_open = 1).  If
        it is 2, don't log to disk at all.

        """
        assert only_terminal in (0, 1, 2)
        if (only_terminal == 0 or (only_terminal == 1 and self.log_file_open)):
            logging_func = self.__call__
        else:
            logging_func = self.log_to_term
            if verbosity >= self.term_verbosity:
                return

        exception_string = self.exception_to_string()
        try:
            logging_func(exception_string, verbosity)
        except IOError:
            print("IOError while trying to log exception!")
            print(exception_string)


Log = Logger()


class ErrorLog:
    """Log each recoverable error in error_log file

    There are three types of recoverable errors:  ListError, which
    happens trying to list a directory or stat a file, UpdateError,
    which happen when trying to update a changed file, and
    SpecialFileError, which happen when a special file cannot be
    created.  See the error policy file for more info.

    """
    _log_fileobj = None

    @classmethod
    def open(cls, time_string, compress=1):
        """Open the error log, prepare for writing"""
        if not Globals.isbackup_writer:
            return Globals.backup_writer.log.ErrorLog.open(
                time_string, compress)
        assert not cls._log_fileobj, "log already open"
        assert Globals.isbackup_writer

        base_rp = Globals.rbdir.append("error_log.%s.data" % time_string)
        if compress:
            cls._log_fileobj = rpath.MaybeGzip(base_rp)
        else:
            cls._log_fileobj = base_rp.open("wb", compress=0)

    @classmethod
    def isopen(cls):
        """True if the error log file is currently open"""
        if Globals.isbackup_writer or not Globals.backup_writer:
            return cls._log_fileobj is not None
        else:
            return Globals.backup_writer.log.ErrorLog.isopen()

    @classmethod
    def write(cls, error_type, rp, exc):
        """Add line to log file indicating error exc with file rp"""
        if not Globals.isbackup_writer:
            return Globals.backup_writer.log.ErrorLog.write(
                error_type, rp, exc)
        logstr = cls.get_log_string(error_type, rp, exc)
        Log(logstr, 2)
        if Globals.null_separator:
            logstr += "\0"
        else:
            logstr = re.sub("\n", " ", logstr)
            logstr += "\n"
        cls._log_fileobj.write(_to_bytes(logstr))

    @classmethod
    def get_indexpath(cls, obj):
        """Return filename for logging.  rp is a rpath, string, or tuple"""
        if getattr(obj, 'get_safeindexpath', None):
            return obj.get_safeindexpath()
        else:
            return repr(obj)

    @classmethod
    def write_if_open(cls, error_type, rp, exc):
        """Call cls.write(...) if error log open, only log otherwise"""
        if not Globals.isbackup_writer and Globals.backup_writer:
            return Globals.backup_writer.log.ErrorLog.write_if_open(
                error_type, rp, exc)
        if cls.isopen():
            cls.write(error_type, rp, exc)
        else:
            Log(cls.get_log_string(error_type, rp, exc), 2)

    @classmethod
    def get_log_string(cls, error_type, rp, exc):
        """Return log string to put in error log"""
        assert (error_type == "ListError" or error_type == "UpdateError"
                or error_type == "SpecialFileError"), "Unknown type " + error_type
        return "%s: '%s' %s" % (error_type, cls.get_indexpath(rp), exc)

    @classmethod
    def close(cls):
        """Close the error log file"""
        if not Globals.isbackup_writer:
            return Globals.backup_writer.log.ErrorLog.close()
        assert not cls._log_fileobj.close()
        cls._log_fileobj = None



# Copyright 2003 Ben Escoto
#
# This file is part of rdiff-backup.
#
# rdiff-backup is free software; you can redistribute it and/or modify
# under the terms of the GNU General Public License as published by the
# Free Software Foundation; either version 2 of the License, or (at your
# option) any later version.
#
# rdiff-backup is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with rdiff-backup; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA
# 02110-1301, USA
"""Determine the capabilities of given file system

rdiff-backup needs to read and write to file systems with varying
abilities.  For instance, some file systems and not others have ACLs,
are case-sensitive, or can store ownership information.  The code in
this module tests the file system for various features, and returns an
FSAbilities object describing it.

"""

import errno
import os
from . import Globals, log, TempFile, selection, robust, SetConnections, \
    FilenameMapping, win_acls, Time


class FSAbilities:
    """Store capabilities of given file system"""
    extended_filenames = None  # True if filenames can have non-ASCII chars
    win_reserved_filenames = None  # True if filenames can't have ",*,: etc.
    case_sensitive = None  # True if "foobar" and "FoObAr" are different files
    ownership = None  # True if chown works on this filesystem
    acls = None  # True if access control lists supported
    eas = None  # True if extended attributes supported
    win_acls = None  # True if windows access control lists supported
    hardlinks = None  # True if hard linking supported
    fsync_dirs = None  # True if directories can be fsync'd
    dir_inc_perms = None  # True if regular files can have full permissions
    resource_forks = None  # True if system supports resource forks
    carbonfile = None  # True if Mac Carbon file data is supported.
    name = None  # Short string, not used for any technical purpose
    read_only = None  # True if capabilities were determined non-destructively
    high_perms = None  # True if suid etc perms are (read/write) supported
    escape_dos_devices = None  # True if dos device files can't be created (e.g.,
    # aux, con, com1, etc)
    escape_trailing_spaces = None  # True if trailing spaces or periods at the
    # end of filenames aren't preserved
    symlink_perms = None  # True if symlink perms are affected by umask

    def __init__(self, name=None):
        """FSAbilities initializer.  name is only used in logging"""
        self.name = name

    def __str__(self):
        """Return pretty printable version of self"""
        assert self.read_only == 0 or self.read_only == 1, self.read_only
        s = ['-' * 65]

        def addline(desc, val_text):
            """Add description line to s"""
            s.append('  %s%s%s' % (desc, ' ' * (45 - len(desc)), val_text))

        def add_boolean_list(pair_list):
            """Add lines from list of (desc, boolean) pairs"""
            for desc, boolean in pair_list:
                if boolean:
                    val_text = 'On'
                elif boolean is None:
                    val_text = 'N/A'
                else:
                    assert boolean == 0
                    val_text = 'Off'
                addline(desc, val_text)

        def get_title_line():
            """Add the first line, mostly for decoration"""
            read_string = self.read_only and "read only" or "read/write"
            if self.name:
                return ('Detected abilities for %s (%s) file system:' %
                        (self.name, read_string))
            else:
                return (
                    'Detected abilities for %s file system' % (read_string, ))

        s.append(get_title_line())
        if not self.read_only:
            add_boolean_list([
                ('Ownership changing', self.ownership),
                ('Hard linking', self.hardlinks),
                ('fsync() directories', self.fsync_dirs),
                ('Directory inc permissions', self.dir_inc_perms),
                ('High-bit permissions', self.high_perms),
                ('Symlink permissions', self.symlink_perms),
                ('Extended filenames', self.extended_filenames),
                ('Windows reserved filenames', self.win_reserved_filenames),
            ])
        add_boolean_list(
            [('Access control lists', self.acls),
             ('Extended attributes', self.eas),
             ('Windows access control lists', self.win_acls),
             ('Case sensitivity', self.case_sensitive),
             ('Escape DOS devices', self.escape_dos_devices),
             ('Escape trailing spaces', self.escape_trailing_spaces),
             ('Mac OS X style resource forks', self.resource_forks),
             ('Mac OS X Finder information', self.carbonfile)])
        s.append(s[0])
        return '\n'.join(s)

    def init_readonly(self, rp):
        """Set variables using fs tested at RPath rp.  Run locally.

        This method does not write to the file system at all, and
        should be run on the file system when the file system will
        only need to be read.

        Only self.acls and self.eas are set.

        """
        assert rp.conn is Globals.local_connection
        self.root_rp = rp
        self.read_only = 1
        self.set_eas(rp, 0)
        self.set_acls(rp)
        self.set_win_acls(rp, 0)
        self.set_resource_fork_readonly(rp)
        self.set_carbonfile()
        self.set_case_sensitive_readonly(rp)
        self.set_escape_dos_devices(rp)
        self.set_escape_trailing_spaces_readonly(rp)
        return self

    def init_readwrite(self, rbdir):
        """Set variables using fs tested at rp_base.  Run locally.

        This method creates a temp directory in rp_base and writes to
        it in order to test various features.  Use on a file system
        that will be written to.

        """
        assert rbdir.conn is Globals.local_connection
        if not rbdir.isdir():
            assert not rbdir.lstat(), (rbdir.path, rbdir.lstat())
            rbdir.mkdir()
        self.root_rp = rbdir
        self.read_only = 0
        subdir = TempFile.new_in_dir(rbdir)
        subdir.mkdir()

        self.set_extended_filenames(subdir)
        self.set_win_reserved_filenames(subdir)
        self.set_case_sensitive_readwrite(subdir)
        self.set_ownership(subdir)
        self.set_hardlinks(subdir)
        self.set_fsync_dirs(subdir)
        self.set_eas(subdir, 1)
        self.set_acls(subdir)
        self.set_win_acls(subdir, 1)
        self.set_dir_inc_perms(subdir)
        self.set_resource_fork_readwrite(subdir)
        self.set_carbonfile()
        self.set_high_perms_readwrite(subdir)
        self.set_symlink_perms(subdir)
        self.set_escape_dos_devices(subdir)
        self.set_escape_trailing_spaces_readwrite(subdir)

        subdir.delete()
        return self

    def set_ownership(self, testdir):
        """Set self.ownership to true iff testdir's ownership can be changed"""
        tmp_rp = testdir.append("foo")
        tmp_rp.touch()
        uid, gid = tmp_rp.getuidgid()
        try:
            tmp_rp.chown(uid // 2 + 1, gid // 2 + 1)  # just choose random uid/gid
            tmp_rp.chown(0, 0)
        except (IOError, OSError, AttributeError):
            self.ownership = 0
        else:
            self.ownership = 1
        tmp_rp.delete()

    def set_hardlinks(self, testdir):
        """Set self.hardlinks to true iff hard linked files can be made"""
        hl_source = testdir.append("hardlinked_file1")
        hl_dir = testdir.append("hl")
        hl_dir.mkdir()
        hl_dest = hl_dir.append("hardlinked_file2")
        hl_source.touch()
        try:
            hl_dest.hardlink(hl_source.path)
            if hl_source.getinode() != hl_dest.getinode():
                raise IOError(errno.EOPNOTSUPP, "Hard links don't compare")
        except (IOError, OSError, AttributeError):
            if Globals.preserve_hardlinks != 0:
                log.Log(
                    "Warning: hard linking not supported by filesystem "
                    "at %s" % self.root_rp.get_safepath(), 3)
            self.hardlinks = None
        else:
            self.hardlinks = 1

    def set_fsync_dirs(self, testdir):
        """Set self.fsync_dirs if directories can be fsync'd"""
        assert testdir.conn is Globals.local_connection
        try:
            testdir.fsync()
        except (IOError, OSError):
            log.Log(
                "Directories on file system at %s are not fsyncable.\n"
                "Assuming it's unnecessary." % testdir.get_safepath(), 4)
            self.fsync_dirs = 0
        else:
            self.fsync_dirs = 1

    def set_extended_filenames(self, subdir):
        """Set self.extended_filenames by trying to write a path"""
        assert not self.read_only

        # Make sure ordinary filenames ok
        ordinary_filename = b'5-_ a.snapshot.gz'
        ord_rp = subdir.append(ordinary_filename)
        ord_rp.touch()
        assert ord_rp.lstat()
        ord_rp.delete()

        # Try path with UTF-8 encoded character
        extended_filename = (
            'uni' + chr(225) + chr(132) + chr(137)).encode('utf-8')
        ext_rp = None
        try:
            ext_rp = subdir.append(extended_filename)
            ext_rp.touch()
        except (IOError, OSError):
            if ext_rp:
                assert not ext_rp.lstat()
            self.extended_filenames = 0
        else:
            assert ext_rp.lstat()
            try:
                ext_rp.delete()
            except (IOError, OSError):
                # Broken CIFS setups will sometimes create UTF-8 files
                # and even stat them, but not let us perform file operations
                # on them. Test file cannot be deleted. UTF-8 chars not in the
                # underlying codepage get translated to '?'
                log.Log.FatalError(
                    "Could not delete extended filenames test "
                    "file. If you are using a CIFS share, please"
                    " see the FAQ entry about characters being "
                    "transformed to a '?'")
            self.extended_filenames = 1

    def set_win_reserved_filenames(self, subdir):
        """Set self.win_reserved_filenames by trying to write a path"""
        assert not self.read_only

        # Try Windows reserved characters
        win_reserved_filename = ':\\"'
        win_rp = None
        try:
            win_rp = subdir.append(win_reserved_filename)
            win_rp.touch()
        except (IOError, OSError):
            if win_rp:
                assert not win_rp.lstat()
            self.win_reserved_filenames = 1
        else:
            assert win_rp.lstat()
            try:
                win_rp.delete()
            except (IOError, OSError):
                self.win_reserved_filenames = 1
            else:
                self.win_reserved_filenames = 0

    def set_acls(self, rp):
        """Set self.acls based on rp.  Does not write.  Needs to be local"""
        assert Globals.local_connection is rp.conn
        assert rp.lstat()
        if Globals.acls_active == 0:
            log.Log(
                "POSIX ACLs test skipped. rdiff-backup run "
                "with --no-acls option.", 4)
            self.acls = 0
            return

        try:
            import posix1e
        except ImportError:
            log.Log(
                "Unable to import module posix1e from pylibacl "
                "package.\nPOSIX ACLs not supported on filesystem at %s" %
                rp.get_safepath(), 4)
            self.acls = 0
            return

        try:
            posix1e.ACL(file=rp.path)
        except IOError:
            log.Log(
                "POSIX ACLs not supported by filesystem at %s" %
                rp.get_safepath(), 4)
            self.acls = 0
        else:
            self.acls = 1

    def set_case_sensitive_readwrite(self, subdir):
        """Determine if directory at rp is case sensitive by writing"""
        assert not self.read_only
        upper_a = subdir.append("A")
        upper_a.touch()
        lower_a = subdir.append("a")
        if lower_a.lstat():
            lower_a.delete()
            upper_a.setdata()
            if upper_a.lstat():
                # we know that (fuse-)exFAT 1.3.0 takes 1sec to register the
                # deletion (July 2020)
                log.Log.FatalError(
                    "We're sorry but the target file system at '%s' isn't "
                    "deemed reliable enough for a backup. It takes too long "
                    "or doesn't register case insensitive deletion of files."
                    % subdir.get_safepath())
            self.case_sensitive = 0
        else:
            upper_a.delete()
            self.case_sensitive = 1

    def set_case_sensitive_readonly(self, rp):
        """Determine if directory at rp is case sensitive without writing"""

        def find_letter(subdir):
            """Find a (subdir_rp, dirlist) with a letter in it, or None

            Recurse down the directory, looking for any file that has
            a letter in it.  Return the pair (rp, [list of filenames])
            where the list is of the directory containing rp.

            """
            files_list = robust.listrp(subdir)
            for filename in files_list:
                if filename != filename.swapcase():
                    return (subdir, files_list, filename)
            for filename in files_list:
                dir_rp = subdir.append(filename)
                if dir_rp.isdir():
                    subsearch = find_letter(dir_rp)
                    if subsearch:
                        return subsearch
            return None

        def test_triple(dir_rp, dirlist, filename):
            """Return 1 if filename shows system case sensitive"""
            try:
                letter_rp = dir_rp.append(filename)
            except OSError:
                return 0
            assert letter_rp.lstat(), letter_rp
            swapped = filename.swapcase()
            if swapped in dirlist:
                return 1

            swapped_rp = dir_rp.append(swapped)
            if swapped_rp.lstat():
                return 0
            return 1

        triple = find_letter(rp)
        if not triple:
            log.Log(
                "Warning: could not determine case sensitivity of "
                "source directory at\n  %s\n"
                "because we can't find any files with letters in them.\n"
                "It will be treated as case sensitive." % rp.get_safepath(), 2)
            self.case_sensitive = 1
            return

        self.case_sensitive = test_triple(*triple)

    def set_eas(self, rp, write):
        """Set extended attributes from rp. Tests writing if write is true."""
        assert Globals.local_connection is rp.conn
        assert rp.lstat()
        if Globals.eas_active == 0:
            log.Log(
                "Extended attributes test skipped. rdiff-backup run "
                "with --no-eas option.", 4)
            self.eas = 0
            return
        try:
            import xattr.pyxattr_compat as xattr
        except ImportError:
            try:
                import xattr
            except ImportError:
                log.Log(
                    "Unable to import module (py)xattr.\nExtended attributes not "
                    "supported on filesystem at %s" % (rp.get_safepath(), ), 4)
                self.eas = 0
                return

        try:
            xattr.list(rp.path)
            if write:
                xattr.set(rp.path, b"user.test", b"test val")
                assert xattr.get(rp.path, b"user.test") == b"test val"
        except IOError:
            log.Log(
                "Extended attributes not supported by "
                "filesystem at %s" % (rp.get_safepath(), ), 4)
            self.eas = 0
        except AssertionError:
            log.Log(
                "Extended attributes support is broken on filesystem at "
                "%s.\nPlease upgrade the filesystem driver, contact the "
                "developers,\nor use the --no-eas option to disable "
                "extended attributes\nsupport and suppress this message." %
                (rp.get_safepath(), ), 1)
            self.eas = 0
        else:
            self.eas = 1

    def set_win_acls(self, dir_rp, write):
        """Test if windows access control lists are supported"""
        assert Globals.local_connection is dir_rp.conn
        assert dir_rp.lstat()
        if Globals.win_acls_active == 0:
            log.Log(
                "Windows ACLs test skipped. rdiff-backup run "
                "with --no-acls option.", 4)
            self.win_acls = 0
            return

        try:
            import win32security
            import pywintypes
        except ImportError:
            log.Log(
                "Unable to import win32security module. Windows ACLs\n"
                "not supported by filesystem at %s" % dir_rp.get_safepath(), 4)
            self.win_acls = 0
            return
        try:
            sd = win32security.GetNamedSecurityInfo(
                os.fsdecode(dir_rp.path), win32security.SE_FILE_OBJECT,
                win32security.OWNER_SECURITY_INFORMATION
                | win32security.GROUP_SECURITY_INFORMATION
                | win32security.DACL_SECURITY_INFORMATION)
            acl = sd.GetSecurityDescriptorDacl()
            acl.GetAceCount()  # to verify that it works
            if write:
                win32security.SetNamedSecurityInfo(
                    os.fsdecode(dir_rp.path), win32security.SE_FILE_OBJECT,
                    win32security.OWNER_SECURITY_INFORMATION
                    | win32security.GROUP_SECURITY_INFORMATION
                    | win32security.DACL_SECURITY_INFORMATION,
                    sd.GetSecurityDescriptorOwner(),
                    sd.GetSecurityDescriptorGroup(),
                    sd.GetSecurityDescriptorDacl(), None)
        except (OSError, AttributeError, pywintypes.error):
            log.Log(
                "Unable to load a Windows ACL.\nWindows ACLs not supported "
                "by filesystem at %s" % dir_rp.get_safepath(), 4)
            self.win_acls = 0
            return

        try:
            win_acls.init_acls()
        except (OSError, AttributeError, pywintypes.error):
            log.Log(
                "Unable to init win_acls.\nWindows ACLs not supported by "
                "filesystem at %s" % dir_rp.get_safepath(), 4)
            self.win_acls = 0
            return
        self.win_acls = 1

    def set_dir_inc_perms(self, rp):
        """See if increments can have full permissions like a directory"""
        test_rp = rp.append('dir_inc_check')
        test_rp.touch()
        try:
            test_rp.chmod(0o7777, 4)
        except OSError:
            test_rp.delete()
            self.dir_inc_perms = 0
            return
        test_rp.setdata()
        assert test_rp.isreg()
        if test_rp.getperms() == 0o7777 or test_rp.getperms() == 0o6777:
            self.dir_inc_perms = 1
        else:
            self.dir_inc_perms = 0
        test_rp.delete()

    def set_carbonfile(self):
        """Test for support of the Mac Carbon library.  This library
        can be used to obtain Finder info (creator/type)."""
        try:
            import Carbon.File
        except (ImportError, AttributeError):
            self.carbonfile = 0
            return

        try:
            Carbon.File.FSSpec('.')  # just to verify that it works
        except BaseException:
            self.carbonfile = 0
            return

        self.carbonfile = 1

    def set_resource_fork_readwrite(self, dir_rp):
        """Test for resource forks by writing to regular_file/..namedfork/rsrc"""
        assert dir_rp.conn is Globals.local_connection
        reg_rp = dir_rp.append('regfile')
        reg_rp.touch()

        s = b'test string---this should end up in resource fork'
        try:
            fp_write = open(
                os.path.join(reg_rp.path, b'..namedfork', b'rsrc'), 'wb')
            fp_write.write(s)
            assert not fp_write.close()

            fp_read = open(
                os.path.join(reg_rp.path, b'..namedfork', b'rsrc'), 'rb')
            s_back = fp_read.read()
            assert not fp_read.close()
        except (OSError, IOError):
            self.resource_forks = 0
        else:
            self.resource_forks = (s_back == s)
        reg_rp.delete()

    def set_resource_fork_readonly(self, dir_rp):
        """Test for resource fork support by testing an regular file

        Launches search for regular file in given directory.  If no
        regular file is found, resource_fork support will be turned
        off by default.

        """
        for rp in selection.Select(dir_rp).set_iter():
            if rp.isreg():
                try:
                    rfork = rp.append(b'..namedfork', b'rsrc')
                    fp = rfork.open('rb')
                    fp.read()
                    assert not fp.close()
                except (OSError, IOError):
                    self.resource_forks = 0
                    return
                self.resource_forks = 1
                return
        self.resource_forks = 0

    def set_high_perms_readwrite(self, dir_rp):
        """Test for writing high-bit permissions like suid"""
        tmpf_rp = dir_rp.append(b"high_perms_file")
        tmpf_rp.touch()
        tmpd_rp = dir_rp.append(b"high_perms_dir")
        tmpd_rp.touch()
        try:
            tmpf_rp.chmod(0o7000, 4)
            tmpf_rp.chmod(0o7777, 4)
            tmpd_rp.chmod(0o7000, 4)
            tmpd_rp.chmod(0o7777, 4)
        except (OSError, IOError):
            self.high_perms = 0
        else:
            self.high_perms = 1
        tmpf_rp.delete()
        tmpd_rp.delete()

    def set_symlink_perms(self, dir_rp):
        """Test if symlink permissions are affected by umask"""
        sym_source = dir_rp.append(b"symlinked_file1")
        sym_source.touch()
        sym_dest = dir_rp.append(b"symlinked_file2")
        try:
            sym_dest.symlink(b"symlinked_file1")
        except (OSError, AttributeError):
            self.symlink_perms = 0
        else:
            sym_dest.setdata()
            assert sym_dest.issym()
            if sym_dest.getperms() == 0o700:
                self.symlink_perms = 1
            else:
                self.symlink_perms = 0
            sym_dest.delete()
        sym_source.delete()

    def set_escape_dos_devices(self, subdir):
        """Test if DOS device files can be used as filenames.

        This test must detect if the underlying OS is Windows, whether we are
        running under Cygwin or natively. Cygwin allows these special files to
        be stat'd from any directory. Native Windows returns OSError (like
        non-Cygwin POSIX), but we can check for that using os.name.

        Note that 'con' and 'aux' have some unusual behaviors as shown below.

        os.lstat()   |  con         aux         prn
        -------------+-------------------------------------
        Unix         |  OSError,2   OSError,2   OSError,2
        Cygwin/NTFS  |  -success-   -success-   -success-
        Cygwin/FAT32 |  -success-   -HANGS-
        Native Win   |  WinError,2  WinError,87 WinError,87
        """
        if os.name == "nt":
            self.escape_dos_devices = 1
            return

        try:
            device_rp = subdir.append(b"con")
            if device_rp.lstat():
                self.escape_dos_devices = 1
            else:
                self.escape_dos_devices = 0
        except (OSError):
            self.escape_dos_devices = 1

    def set_escape_trailing_spaces_readwrite(self, testdir):
        """
        Windows and Linux/FAT32 will not preserve trailing spaces or periods.
        Linux/FAT32 behaves inconsistently: It will give an OSError,22 if
        os.mkdir() is called on a directory name with a space at the end, but
        will give an IOError("invalid mode") if you attempt to create a filename
        with a space at the end. However, if a period is placed at the end of
        the name, Linux/FAT32 is consistent with Cygwin and Native Windows.
        """

        period_rp = testdir.append("foo.")
        assert not period_rp.lstat()

        tmp_rp = testdir.append("foo")
        tmp_rp.touch()
        assert tmp_rp.lstat()

        period_rp.setdata()
        if period_rp.lstat():
            self.escape_trailing_spaces = 1
        else:
            self.escape_trailing_spaces = 0

        tmp_rp.delete()

    def set_escape_trailing_spaces_readonly(self, rp):
        """Determine if directory at rp permits filenames with trailing
        spaces or periods without writing."""

        def test_period(dir_rp, dirlist):
            """Return 1 if trailing spaces and periods should be escaped"""
            filename = dirlist[0]
            try:
                test_rp = dir_rp.append(filename)
            except OSError:
                return 0
            assert test_rp.lstat(), test_rp
            period = filename + b'.'
            if period in dirlist:
                return 0

            return 0  # FIXME the following lines fail if filename is almost too long
            period_rp = dir_rp.append(period)
            if period_rp.lstat():
                return 1
            return 0

        dirlist = robust.listrp(rp)
        if len(dirlist):
            self.escape_trailing_spaces = test_period(rp, dirlist)
        else:
            log.Log(
                "Warning: could not determine if source directory at\n"
                "  %s\npermits trailing spaces or periods in "
                "filenames because we can't find any files.\n"
                "It will be treated as permitting such files." %
                rp.get_safepath(), 2)
            self.escape_trailing_spaces = 0


def get_readonly_fsa(desc_string, rp):
    """Return an fsa with given description_string

    Will be initialized read_only with given RPath rp.  We separate
    this out into a separate function so the request can be vetted by
    the security module.

    """
    if os.name == 'nt':
        log.Log("Hardlinks disabled by default on Windows", 4)
        SetConnections.UpdateGlobal('preserve_hardlinks', 0)
    return FSAbilities(desc_string).init_readonly(rp)


class SetGlobals:
    """Various functions for setting Globals vars given FSAbilities above

    Container for BackupSetGlobals and RestoreSetGlobals (don't use directly)

    """

    def __init__(self, in_conn, out_conn, src_fsa, dest_fsa):
        """Just store some variables for use below"""
        self.in_conn, self.out_conn = in_conn, out_conn
        self.src_fsa, self.dest_fsa = src_fsa, dest_fsa

    def set_eas(self):
        self.update_triple(self.src_fsa.eas, self.dest_fsa.eas,
                           ('eas_active', 'eas_write', 'eas_conn'))

    def set_acls(self):
        self.update_triple(self.src_fsa.acls, self.dest_fsa.acls,
                           ('acls_active', 'acls_write', 'acls_conn'))
        if Globals.never_drop_acls and not Globals.acls_active:
            log.Log.FatalError("--never-drop-acls specified, but ACL support\n"
                               "missing from source filesystem")

    def set_win_acls(self):
        self.update_triple(
            self.src_fsa.win_acls, self.dest_fsa.win_acls,
            ('win_acls_active', 'win_acls_write', 'win_acls_conn'))

    def set_resource_forks(self):
        self.update_triple(self.src_fsa.resource_forks,
                           self.dest_fsa.resource_forks,
                           ('resource_forks_active', 'resource_forks_write',
                            'resource_forks_conn'))

    def set_carbonfile(self):
        self.update_triple(
            self.src_fsa.carbonfile, self.dest_fsa.carbonfile,
            ('carbonfile_active', 'carbonfile_write', 'carbonfile_conn'))

    def set_hardlinks(self):
        if Globals.preserve_hardlinks != 0:
            SetConnections.UpdateGlobal('preserve_hardlinks',
                                        self.dest_fsa.hardlinks)

    def set_fsync_directories(self):
        SetConnections.UpdateGlobal('fsync_directories',
                                    self.dest_fsa.fsync_dirs)

    def set_change_ownership(self):
        SetConnections.UpdateGlobal('change_ownership',
                                    self.dest_fsa.ownership)

    def set_high_perms(self):
        if not self.dest_fsa.high_perms:
            SetConnections.UpdateGlobal('permission_mask', 0o777)

    def set_symlink_perms(self):
        SetConnections.UpdateGlobal('symlink_perms',
                                    self.dest_fsa.symlink_perms)

    def set_compatible_timestamps(self):
        if Globals.chars_to_quote.find(b":") > -1:
            SetConnections.UpdateGlobal('use_compatible_timestamps', 1)
            Time.setcurtime(
                Time.curtime)  # update Time.curtimestr on all conns
            log.Log("Enabled use_compatible_timestamps", 4)


class BackupSetGlobals(SetGlobals):
    """Functions for setting fsa related globals for backup session"""

    def update_triple(self, src_support, dest_support, attr_triple):
        """Many of the settings have a common form we can handle here"""
        active_attr, write_attr, conn_attr = attr_triple
        if Globals.get(active_attr) == 0:
            return  # don't override 0
        for attr in attr_triple:
            SetConnections.UpdateGlobal(attr, None)
        if not src_support:
            return  # if source doesn't support, nothing
        SetConnections.UpdateGlobal(active_attr, 1)
        self.in_conn.Globals.set_local(conn_attr, 1)
        if dest_support:
            SetConnections.UpdateGlobal(write_attr, 1)
            self.out_conn.Globals.set_local(conn_attr, 1)

    def set_special_escapes(self, rbdir):
        """Escaping DOS devices and trailing periods/spaces works like
        regular filename escaping. If only the destination requires it,
        then we do it. Otherwise, it is not necessary, since the files
        couldn't have been created in the first place. We also record
        whether we have done it in order to handle the case where a
        volume which was escaped is later restored by an OS that does
        not require it.

        """

        suggested_edd = (self.dest_fsa.escape_dos_devices
                         and not self.src_fsa.escape_dos_devices)
        suggested_ets = (self.dest_fsa.escape_trailing_spaces
                         and not self.src_fsa.escape_trailing_spaces)

        se_rp = rbdir.append("special_escapes")
        if not se_rp.lstat():
            actual_edd, actual_ets = suggested_edd, suggested_ets
            se = ""
            if actual_edd:
                se = se + "escape_dos_devices\n"
            if actual_ets:
                se = se + "escape_trailing_spaces\n"
            se_rp.write_string(se)
        else:
            se = se_rp.get_string().split("\n")
            actual_edd = ("escape_dos_devices" in se)
            actual_ets = ("escape_trailing_spaces" in se)

            if actual_edd != suggested_edd and not suggested_edd:
                log.Log(
                    "Warning: System no longer needs DOS devices escaped, "
                    "but we will retain for backwards compatibility.", 2)
            if actual_ets != suggested_ets and not suggested_ets:
                log.Log(
                    "Warning: System no longer needs trailing spaces or "
                    "periods escaped, but we will retain for backwards "
                    "compatibility.", 2)

        SetConnections.UpdateGlobal('escape_dos_devices', actual_edd)
        log.Log("Backup: escape_dos_devices = %d" % actual_edd, 4)

        SetConnections.UpdateGlobal('escape_trailing_spaces', actual_ets)
        log.Log("Backup: escape_trailing_spaces = %d" % actual_ets, 4)

    def set_chars_to_quote(self, rbdir, force):
        """Set chars_to_quote setting for backup session

        Unlike most other options, the chars_to_quote setting also
        depends on the current settings in the rdiff-backup-data
        directory, not just the current fs features.

        """
        (ctq, update) = self.compare_ctq_file(rbdir, self.get_ctq_from_fsas(),
                                              force)

        SetConnections.UpdateGlobal('chars_to_quote', ctq)
        if Globals.chars_to_quote:
            FilenameMapping.set_init_quote_vals()
        return update

    def get_ctq_from_fsas(self):
        """Determine chars_to_quote just from filesystems, no ctq file"""
        ctq = []

        if self.src_fsa.case_sensitive and not self.dest_fsa.case_sensitive:
            ctq.append(b"A-Z")  # Quote upper case
        if not self.dest_fsa.extended_filenames:
            ctq.append(b'\000-\037')  # Quote 0 - 31
            ctq.append(b'\200-\377')  # Quote non-ASCII characters 0x80 - 0xFF
        if self.dest_fsa.win_reserved_filenames:
            if self.dest_fsa.extended_filenames:
                ctq.append(b'\000-\037')  # Quote 0 - 31
            # Quote ", *, /, :, <, >, ?, \, |, and 127 (DEL)
            ctq.append(b'\"*/:<>?\\\\|\177')

        # Quote quoting char if quoting anything
        if ctq:
            ctq.append(Globals.quoting_char)
        return b"".join(ctq)

    def compare_ctq_file(self, rbdir, suggested_ctq, force):
        """Compare ctq file with suggested result, return actual ctq"""
        ctq_rp = rbdir.append(b"chars_to_quote")
        if not ctq_rp.lstat():
            if Globals.chars_to_quote is None:
                actual_ctq = suggested_ctq
            else:
                actual_ctq = Globals.chars_to_quote
            ctq_rp.write_bytes(actual_ctq)
            return (actual_ctq, None)

        if Globals.chars_to_quote is None:
            actual_ctq = ctq_rp.get_bytes()
        else:
            actual_ctq = Globals.chars_to_quote  # Globals override

        if actual_ctq == suggested_ctq:
            return (actual_ctq, None)
        if suggested_ctq == b"":
            log.Log(
                "Warning: File system no longer needs quoting, "
                "but we will retain for backwards compatibility.", 2)
            return (actual_ctq, None)
        if Globals.chars_to_quote is None:
            if force:
                log.Log(
                    "Warning: migrating rdiff-backup repository from"
                    "old quoting chars %r to new quoting chars %r" %
                    (actual_ctq, suggested_ctq), 2)
                ctq_rp.delete()
                ctq_rp.write_bytes(suggested_ctq)
                return (suggested_ctq, 1)
            else:
                log.Log.FatalError(
                    """New quoting requirements!

The quoting chars this session needs %r do not match
the repository settings %r listed in

%s

This may be caused when you copy an rdiff-backup repository from a
normal file system onto a windows one that cannot support the same
characters, or if you backup a case-sensitive file system onto a
case-insensitive one that previously only had case-insensitive ones
backed up onto it.

By specifying the --force option, rdiff-backup will migrate the
repository from the old quoting chars to the new ones.""" %
                    (suggested_ctq, actual_ctq, ctq_rp.get_safepath()))
        return (actual_ctq, None)  # Maintain Globals override


class RestoreSetGlobals(SetGlobals):
    """Functions for setting fsa-related globals for restore session"""

    def update_triple(self, src_support, dest_support, attr_triple):
        """Update global settings for feature based on fsa results

        This is slightly different from BackupSetGlobals.update_triple
        because (using the mirror_metadata file) rpaths from the
        source may have more information than the file system
        supports.

        """
        active_attr, write_attr, conn_attr = attr_triple
        if Globals.get(active_attr) == 0:
            return  # don't override 0
        for attr in attr_triple:
            SetConnections.UpdateGlobal(attr, None)
        if not dest_support:
            return  # if dest doesn't support, do nothing
        SetConnections.UpdateGlobal(active_attr, 1)
        self.out_conn.Globals.set_local(conn_attr, 1)
        self.out_conn.Globals.set_local(write_attr, 1)
        if src_support:
            self.in_conn.Globals.set_local(conn_attr, 1)

    def set_special_escapes(self, rbdir):
        """Set escape_dos_devices and escape_trailing_spaces from
        rdiff-backup-data dir, just like chars_to_quote"""
        se_rp = rbdir.append("special_escapes")
        if se_rp.lstat():
            se = se_rp.get_string().split("\n")
            actual_edd = ("escape_dos_devices" in se)
            actual_ets = ("escape_trailing_spaces" in se)
        else:
            log.Log(
                "Warning: special_escapes file not found,\n"
                "will assume need to escape DOS devices and trailing "
                "spaces based on file systems.", 2)
            if getattr(self, "src_fsa", None) is not None:
                actual_edd = (self.src_fsa.escape_dos_devices
                              and not self.dest_fsa.escape_dos_devices)
                actual_ets = (self.src_fsa.escape_trailing_spaces
                              and not self.dest_fsa.escape_trailing_spaces)
            else:
                # Single filesystem operation
                actual_edd = self.dest_fsa.escape_dos_devices
                actual_ets = self.dest_fsa.escape_trailing_spaces

        SetConnections.UpdateGlobal('escape_dos_devices', actual_edd)
        log.Log("Backup: escape_dos_devices = %d" % actual_edd, 4)

        SetConnections.UpdateGlobal('escape_trailing_spaces', actual_ets)
        log.Log("Backup: escape_trailing_spaces = %d" % actual_ets, 4)

    def set_chars_to_quote(self, rbdir):
        """Set chars_to_quote from rdiff-backup-data dir"""
        if Globals.chars_to_quote is not None:
            return  # already overridden

        ctq_rp = rbdir.append(b"chars_to_quote")
        if ctq_rp.lstat():
            SetConnections.UpdateGlobal("chars_to_quote", ctq_rp.get_bytes())
        else:
            log.Log(
                "Warning: chars_to_quote file not found,\n"
                "assuming no quoting in backup repository.", 2)
            SetConnections.UpdateGlobal("chars_to_quote", b"")


class SingleSetGlobals(RestoreSetGlobals):
    """For setting globals when dealing only with one filesystem"""

    def __init__(self, conn, fsa):
        self.conn = conn
        self.dest_fsa = fsa

    def update_triple(self, fsa_support, attr_triple):
        """Update global vars from single fsa test"""
        active_attr, write_attr, conn_attr = attr_triple
        if Globals.get(active_attr) == 0:
            return  # don't override 0
        for attr in attr_triple:
            SetConnections.UpdateGlobal(attr, None)
        if not fsa_support:
            return
        SetConnections.UpdateGlobal(active_attr, 1)
        SetConnections.UpdateGlobal(write_attr, 1)
        self.conn.Globals.set_local(conn_attr, 1)

    def set_eas(self):
        self.update_triple(self.dest_fsa.eas,
                           ('eas_active', 'eas_write', 'eas_conn'))

    def set_acls(self):
        self.update_triple(self.dest_fsa.acls,
                           ('acls_active', 'acls_write', 'acls_conn'))

    def set_win_acls(self):
        self.update_triple(
            self.src_fsa.win_acls, self.dest_fsa.win_acls,
            ('win_acls_active', 'win_acls_write', 'win_acls_conn'))

    def set_resource_forks(self):
        self.update_triple(self.dest_fsa.resource_forks,
                           ('resource_forks_active', 'resource_forks_write',
                            'resource_forks_conn'))

    def set_carbonfile(self):
        self.update_triple(
            self.dest_fsa.carbonfile,
            ('carbonfile_active', 'carbonfile_write', 'carbonfile_conn'))


def backup_set_globals(rpin, force):
    """Given rps for source filesystem and repository, set fsa globals

    This should be run on the destination connection, because we may
    need to write a new chars_to_quote file.

    """
    assert Globals.rbdir.conn is Globals.local_connection
    src_fsa = rpin.conn.fs_abilities.get_readonly_fsa('source', rpin)
    log.Log(str(src_fsa), 4)
    dest_fsa = FSAbilities('destination').init_readwrite(Globals.rbdir)
    log.Log(str(dest_fsa), 4)

    bsg = BackupSetGlobals(rpin.conn, Globals.rbdir.conn, src_fsa, dest_fsa)
    bsg.set_eas()
    bsg.set_acls()
    bsg.set_win_acls()
    bsg.set_resource_forks()
    bsg.set_carbonfile()
    bsg.set_hardlinks()
    bsg.set_fsync_directories()
    bsg.set_change_ownership()
    bsg.set_high_perms()
    bsg.set_symlink_perms()
    update_quoting = bsg.set_chars_to_quote(Globals.rbdir, force)
    bsg.set_special_escapes(Globals.rbdir)
    bsg.set_compatible_timestamps()

    if update_quoting and force:
        FilenameMapping.update_quoting(Globals.rbdir)


def restore_set_globals(rpout):
    """Set fsa related globals for restore session, given in/out rps"""
    assert rpout.conn is Globals.local_connection
    src_fsa = Globals.rbdir.conn.fs_abilities.get_readonly_fsa(
        'rdiff-backup repository', Globals.rbdir)
    log.Log(str(src_fsa), 4)
    dest_fsa = FSAbilities('restore target').init_readwrite(rpout)
    log.Log(str(dest_fsa), 4)

    rsg = RestoreSetGlobals(Globals.rbdir.conn, rpout.conn, src_fsa, dest_fsa)
    rsg.set_eas()
    rsg.set_acls()
    rsg.set_win_acls()
    rsg.set_resource_forks()
    rsg.set_carbonfile()
    rsg.set_hardlinks()
    # No need to fsync anything when restoring
    rsg.set_change_ownership()
    rsg.set_high_perms()
    rsg.set_symlink_perms()
    rsg.set_chars_to_quote(Globals.rbdir)
    rsg.set_special_escapes(Globals.rbdir)
    rsg.set_compatible_timestamps()


def single_set_globals(rp, read_only=None):
    """Set fsa related globals for operation on single filesystem"""
    if read_only:
        fsa = rp.conn.fs_abilities.get_readonly_fsa(rp.path, rp)
    else:
        fsa = FSAbilities(rp.path).init_readwrite(rp)
    log.Log(str(fsa), 4)

    ssg = SingleSetGlobals(rp.conn, fsa)
    ssg.set_eas()
    ssg.set_acls()
    ssg.set_resource_forks()
    ssg.set_carbonfile()
    if not read_only:
        ssg.set_hardlinks()
        ssg.set_change_ownership()
        ssg.set_high_perms()
        ssg.set_symlink_perms()
    ssg.set_chars_to_quote(Globals.rbdir)
    ssg.set_special_escapes(Globals.rbdir)
    ssg.set_compatible_timestamps()
